from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
import warnings
import pdb
import os

import numpy as np
import warnings

# check transformers version
from transformers import __version__
if __version__ <= '4.19.0':
    warnings.warn("The version of `transformers` is too low, which may cause errors in generation. Please try to run `pip install transformers --upgrade` to upgrade transformers.")
    from transformers.generation_utils import GenerationMixin
    from transformers.generation_utils import (
    SampleOutput, SampleEncoderDecoderOutput,SampleDecoderOnlyOutput
    ,)
    from transformers.file_utils import ModelOutput
    from transformers.utils import logging
    from transformers.generation_stopping_criteria import (
        MaxLengthCriteria,
        MaxTimeCriteria,
        StoppingCriteriaList,
        validate_stopping_criteria,
    )
    from transformers.generation_logits_process import (
        EncoderNoRepeatNGramLogitsProcessor,
        ExponentialDecayLengthPenalty,
        ForcedBOSTokenLogitsProcessor,
        ForcedEOSTokenLogitsProcessor,
        HammingDiversityLogitsProcessor,
        InfNanRemoveLogitsProcessor,
        LogitNormalization,
        LogitsProcessorList,
        MinLengthLogitsProcessor,
        NoBadWordsLogitsProcessor,
        NoRepeatNGramLogitsProcessor,
        PrefixConstrainedLogitsProcessor,
        RepetitionPenaltyLogitsProcessor,
    )
else:
    from transformers.generation.utils import GenerationMixin
    from transformers.generation.utils import (
    SampleOutput, SampleEncoderDecoderOutput,SampleDecoderOnlyOutput
    ,)
    from transformers.file_utils import ModelOutput
    from transformers.utils import logging
    from transformers.generation.stopping_criteria import (
        MaxLengthCriteria,
        MaxTimeCriteria,
        StoppingCriteriaList,
        validate_stopping_criteria,
    )
    from transformers.generation.logits_process import (
        EncoderNoRepeatNGramLogitsProcessor,
        ExponentialDecayLengthPenalty,
        ForcedBOSTokenLogitsProcessor,
        ForcedEOSTokenLogitsProcessor,
        HammingDiversityLogitsProcessor,
        InfNanRemoveLogitsProcessor,
        LogitNormalization,
        LogitsProcessorList,
        MinLengthLogitsProcessor,
        NoBadWordsLogitsProcessor,
        NoRepeatNGramLogitsProcessor,
        PrefixConstrainedLogitsProcessor,
        RepetitionPenaltyLogitsProcessor,
    )

import torch
from torch import nn
import torch.distributed as dist
logger = logging.get_logger(__name__)

class EHRGenerationMixin(GenerationMixin):
    """
    A class containing all of the functions supporting generation, to be used as a mixin in
    :class:`~transformers.PreTrainedModel`.
    """
    def _get_stopping_criteria(self, max_length: Optional[int], max_time: Optional[float]) -> StoppingCriteriaList:
        stopping_criteria = StoppingCriteriaList()
        if max_length is not None:
            stopping_criteria.append(MaxLengthCriteria(max_length=max_length))
        if max_time is not None:
            stopping_criteria.append(MaxTimeCriteria(max_time=max_time))
        return stopping_criteria

    @torch.no_grad()
    def generate(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        max_length: Optional[int] = None,
        min_length: Optional[int] = None,
        do_sample: Optional[bool] = None,
        early_stopping: Optional[bool] = None,
        num_beams: Optional[int] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        bad_words_ids: Optional[Iterable[int]] = None,
        bos_token_id: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        length_penalty: Optional[float] = None,
        no_repeat_ngram_size: Optional[int] = None,
        encoder_no_repeat_ngram_size: Optional[int] = None,
        num_return_sequences: Optional[int] = None,
        max_time: Optional[float] = None,
        max_new_tokens: Optional[int] = None,
        decoder_start_token_id: Optional[int] = None,
        use_cache: Optional[bool] = None,
        num_beam_groups: Optional[int] = None,
        diversity_penalty: Optional[float] = None,
        prefix_allowed_tokens_fn: Optional[Callable[[int, torch.Tensor], List[int]]] = None,
        logits_processor: Optional[LogitsProcessorList] = LogitsProcessorList(),
        renormalize_logits: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        forced_bos_token_id: Optional[int] = None,
        forced_eos_token_id: Optional[int] = None,
        remove_invalid_values: Optional[bool] = None,
        synced_gpus: Optional[bool] = None,
        exponential_decay_length_penalty: Optional[Tuple[Union[int, float]]] = None,
        code_type: Optional[str] = 'diag',
        **model_kwargs,
    ) -> Union[SampleOutput, torch.LongTensor]:
        r'''an adaptation of generate function implemented in transformers.generation_utils.
        '''
        num_beams = num_beams if num_beams is not None else self.config.num_beams
        num_beam_groups = num_beam_groups if num_beam_groups is not None else self.config.num_beam_groups
        do_sample = do_sample if do_sample is not None else self.config.do_sample
        num_return_sequences = (
            num_return_sequences if num_return_sequences is not None else self.config.num_return_sequences
        )

        pad_token_id = pad_token_id if pad_token_id is not None else self.config.pad_token_id
        bos_token_id = bos_token_id if bos_token_id is not None else self.config.bos_token_id
        eos_token_id = eos_token_id if eos_token_id is not None else self.config.eos_token_id

        output_scores = output_scores if output_scores is not None else self.config.output_scores
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict_in_generate = (
            return_dict_in_generate if return_dict_in_generate is not None else self.config.return_dict_in_generate
        )

        model_kwargs["output_attentions"] = output_attentions
        model_kwargs["output_hidden_states"] = output_hidden_states

        if input_ids is None and "inputs_embeds" not in model_kwargs:
            # init `input_ids` with random idxs if not given
            # this is because eos and bos token id is not in the vocabulary for each code
            input_ids = self._prepare_input_ids_for_generation(bos_token_id, model_kwargs.get("encoder_outputs"))

        if model_kwargs.get("attention_mask", None) is None:
            # init `attention_mask` depending on `pad_token_id`
            model_kwargs["attention_mask"] = self._prepare_attention_mask_for_generation(
                input_ids, pad_token_id, eos_token_id
            )

        # special case if pad_token_id is not defined
        if pad_token_id is None and eos_token_id is not None:
            logger.warning(f"Setting `pad_token_id` to `eos_token_id`:{eos_token_id} for open-end generation.")
            pad_token_id = eos_token_id

        # Storing encoder_input_ids for logits_processor that could use them
        encoder_input_ids = input_ids if self.config.is_encoder_decoder else None

        if self.config.is_encoder_decoder: # True            
            # add encoder_outputs to model_kwargs
            model_kwargs = self._prepare_encoder_decoder_kwargs_for_generation(input_ids, model_kwargs)

            # set input_ids as decoder_input_ids
            if "decoder_input_ids" in model_kwargs:
                input_ids = model_kwargs.pop("decoder_input_ids")
            else:
                input_ids = self._prepare_decoder_input_ids_for_generation(
                    batch_size=1, decoder_start_token_id=decoder_start_token_id, bos_token_id=bos_token_id, model_kwargs=model_kwargs,
                )

            if "encoder_outputs" not in model_kwargs or not isinstance(model_kwargs["encoder_outputs"], ModelOutput):
                raise ValueError("Make sure that `model_kwargs` include `encoder_outputs` of type `ModelOutput`.")

        # if `max_new_tokens` is passed, but not `max_length` -> set `max_length = max_new_tokens`
        input_ids_seq_length = input_ids.shape[-1]
        if max_length is None and max_new_tokens is not None:
            max_length = (
                max_new_tokens + input_ids_seq_length
                if input_ids is not None
                else max_length + model_kwargs["inputs_embeds"].shape[1]
            )
        elif max_length is not None and max_new_tokens is not None:
            # Both are set, this is odd, raise a warning
            warnings.warn(
                "Both `max_length` and `max_new_tokens` have been set "
                f"but they serve the same purpose. `max_length` {max_length} "
                f"will take priority over `max_new_tokens` {max_new_tokens}.",
                UserWarning,
            )

        # default to config if still None
        max_length = max_length if max_length is not None else self.config.max_length

        if input_ids.shape[-1] >= max_length:
            input_ids_string = "decoder_input_ids" if self.config.is_encoder_decoder else "input_ids"
            logger.warning(
                f"Input length of {input_ids_string} is {input_ids.shape[-1]}, but ``max_length`` is set to {max_length}. "
                "This can lead to unexpected behavior. You should consider increasing ``config.max_length`` or ``max_length``."
            )

        # determine generation mode
        is_greedy_gen_mode = (num_beams == 1) and (num_beam_groups == 1) and do_sample is False
        is_sample_gen_mode = (num_beams == 1) and (num_beam_groups == 1) and do_sample is True
        is_beam_gen_mode = (num_beams > 1) and (num_beam_groups == 1) and do_sample is False
        is_beam_sample_gen_mode = (num_beams > 1) and (num_beam_groups == 1) and do_sample is True
        is_group_beam_gen_mode = (num_beams > 1) and (num_beam_groups > 1)
        if num_beam_groups > num_beams:
            raise ValueError("`num_beam_groups` has to be smaller or equal to `num_beams`")
        if is_group_beam_gen_mode and do_sample is True:
            raise ValueError(
                "Diverse beam search cannot be used in sampling mode. Make sure that `do_sample` is set to `False`."
            )

        # set model_kwargs
        model_kwargs["use_cache"] = use_cache

        # get distribution pre_processing samplers
        logits_processor = self._get_logits_processor(
            input_ids_seq_length=input_ids_seq_length,
            repetition_penalty=repetition_penalty,
            no_repeat_ngram_size=no_repeat_ngram_size,
            encoder_no_repeat_ngram_size=encoder_no_repeat_ngram_size,
            encoder_input_ids=encoder_input_ids,
            bad_words_ids=bad_words_ids,
            min_length=min_length,
            max_length=max_length,
            eos_token_id=eos_token_id,
            forced_bos_token_id=forced_bos_token_id,
            forced_eos_token_id=forced_eos_token_id,
            prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
            num_beams=num_beams,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            remove_invalid_values=remove_invalid_values,
            exponential_decay_length_penalty=exponential_decay_length_penalty,
            logits_processor=logits_processor,
            renormalize_logits=renormalize_logits,
        )

        stopping_criteria = self._get_stopping_criteria(max_length=max_length, max_time=max_time)

        if is_sample_gen_mode:
            # get probability distribution warper
            logits_warper = self._get_logits_warper(
                top_k=top_k, top_p=top_p, temperature=temperature, num_beams=num_beams
            )

            # expand input_ids with `num_return_sequences` additional sequences per batch
            input_ids, model_kwargs = self._expand_inputs_for_generation(
                input_ids,
                expand_size=num_return_sequences,
                is_encoder_decoder=self.config.is_encoder_decoder,
                **model_kwargs,
            )

            # sample
            return self.sample(
                input_ids,
                logits_processor=logits_processor,
                logits_warper=logits_warper,
                stopping_criteria=stopping_criteria,
                pad_token_id=pad_token_id,
                eos_token_id=eos_token_id,
                output_scores=output_scores,
                return_dict_in_generate=return_dict_in_generate,
                synced_gpus=synced_gpus,
                code_type=code_type,
                **model_kwargs,
            )
        
        else:
            raise ValueError("PromptEHR only supports `is_sample_gen_mode`! Please set `do_sample` to `True")


    def sample(
        self,
        input_ids: torch.LongTensor,
        logits_processor: Optional[LogitsProcessorList] = None,
        stopping_criteria: Optional[StoppingCriteriaList] = None,
        logits_warper: Optional[LogitsProcessorList] = None,
        max_length: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        synced_gpus: Optional[bool] = None,
        code_type: Optional[str] = 'diagnosis',
        **model_kwargs,
    ) -> Union[SampleOutput, torch.LongTensor]:
        r"""
        Generates sequences for models with a language modeling head using multinomial sampling.

        Parameters:

            input_ids (:obj:`torch.LongTensor` of shape :obj:`(batch_size, sequence_length)`):
                The sequence used as a prompt for the generation.
            logits_processor (:obj:`LogitsProcessorList`, `optional`):
                An instance of :class:`~transformers.LogitsProcessorList`. List of instances of class derived from
                :class:`~transformers.LogitsProcessor` used to modify the prediction scores of the language modeling
                head applied at each generation step.
            stopping_criteria (:obj:`StoppingCriteriaList`, `optional`):
                An instance of :class:`~transformers.StoppingCriteriaList`. List of instances of class derived from
                :class:`~transformers.StoppingCriteria` used to tell if the generation loop should stop.
            logits_warper (:obj:`LogitsProcessorList`, `optional`):
                An instance of :class:`~transformers.LogitsProcessorList`. List of instances of class derived from
                :class:`~transformers.LogitsWarper` used to warp the prediction score distribution of the language
                modeling head applied before multinomial sampling at each generation step.
            max_length (:obj:`int`, `optional`, defaults to 20):
                **DEPRECATED**. Use :obj:`logits_processor` or :obj:`stopping_criteria` directly to cap the number of
                generated tokens. The maximum length of the sequence to be generated.
            pad_token_id (:obj:`int`, `optional`):
                The id of the `padding` token.
            eos_token_id (:obj:`int`, `optional`):
                The id of the `end-of-sequence` token.
            output_attentions (:obj:`bool`, `optional`, defaults to `False`):
                Whether or not to return the attentions tensors of all attention layers. See ``attentions`` under
                returned tensors for more details.
            output_hidden_states (:obj:`bool`, `optional`, defaults to `False`):
                Whether or not to return the hidden states of all layers. See ``hidden_states`` under returned tensors
                for more details.
            output_scores (:obj:`bool`, `optional`, defaults to `False`):
                Whether or not to return the prediction scores. See ``scores`` under returned tensors for more details.
            return_dict_in_generate (:obj:`bool`, `optional`, defaults to `False`):
                Whether or not to return a :class:`~transformers.file_utils.ModelOutput` instead of a plain tuple.
            synced_gpus (:obj:`bool`, `optional`, defaults to :obj:`False`):
                Whether to continue running the while loop until max_length (needed for ZeRO stage 3)
            model_kwargs:
                Additional model specific kwargs will be forwarded to the :obj:`forward` function of the model. If
                model is an encoder-decoder model the kwargs should include :obj:`encoder_outputs`.

        Return:
            :class:`~transformers.generation_utils.SampleDecoderOnlyOutput`,
            :class:`~transformers.generation_utils.SampleEncoderDecoderOutput` or obj:`torch.LongTensor`: A
            :obj:`torch.LongTensor` containing the generated tokens (default behaviour) or a
            :class:`~transformers.generation_utils.SampleDecoderOnlyOutput` if
            ``model.config.is_encoder_decoder=False`` and ``return_dict_in_generate=True`` or a
            :class:`~transformers.generation_utils.SampleEncoderDecoderOutput` if
            ``model.config.is_encoder_decoder=True``.

        Examples::

            >>> from transformers import (
            ...    AutoTokenizer,
            ...    AutoModelForCausalLM,
            ...    LogitsProcessorList,
            ...    MinLengthLogitsProcessor,
            ...    TopKLogitsWarper,
            ...    TemperatureLogitsWarper,
            ... )

            >>> tokenizer = AutoTokenizer.from_pretrained("gpt2")
            >>> model = AutoModelForCausalLM.from_pretrained("gpt2")

            >>> # set pad_token_id to eos_token_id because GPT2 does not have a EOS token
            >>> model.config.pad_token_id = model.config.eos_token_id

            >>> input_prompt = "Today is a beautiful day, and"
            >>> input_ids = tokenizer(input_prompt, return_tensors="pt").input_ids

            >>> # instantiate logits processors
            >>> logits_processor = LogitsProcessorList([
            ...     MinLengthLogitsProcessor(15, eos_token_id=model.config.eos_token_id),
            ... ])
            >>> # instantiate logits processors
            >>> logits_warper = LogitsProcessorList([
            ...     TopKLogitsWarper(50),
            ...     TemperatureLogitsWarper(0.7),
            ... ])

            >>> outputs = model.sample(input_ids, logits_processor=logits_processor, logits_warper=logits_warper)

            >>> print("Generated:", tokenizer.batch_decode(outputs, skip_special_tokens=True))
        """

        # init values
        logits_processor = logits_processor if logits_processor is not None else LogitsProcessorList()
        stopping_criteria = stopping_criteria if stopping_criteria is not None else StoppingCriteriaList()
        if max_length is not None:
            warnings.warn(
                "`max_length` is deprecated in this function, use `stopping_criteria=StoppingCriteriaList(MaxLengthCriteria(max_length=max_length))` instead.",
                UserWarning,
            )
            stopping_criteria = validate_stopping_criteria(stopping_criteria, max_length)
        logits_warper = logits_warper if logits_warper is not None else LogitsProcessorList()
        pad_token_id = pad_token_id if pad_token_id is not None else self.config.pad_token_id
        eos_token_id = eos_token_id if eos_token_id is not None else self.config.eos_token_id
        output_scores = output_scores if output_scores is not None else self.config.output_scores
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict_in_generate = (
            return_dict_in_generate if return_dict_in_generate is not None else self.config.return_dict_in_generate
        )

        # init attention / hidden states / scores tuples
        scores = () if (return_dict_in_generate and output_scores) else None
        decoder_attentions = () if (return_dict_in_generate and output_attentions) else None
        cross_attentions = () if (return_dict_in_generate and output_attentions) else None
        decoder_hidden_states = () if (return_dict_in_generate and output_hidden_states) else None

        # if model is an encoder-decoder, retrieve encoder attention weights and hidden states
        if return_dict_in_generate and self.config.is_encoder_decoder:
            encoder_attentions = model_kwargs["encoder_outputs"].get("attentions") if output_attentions else None
            encoder_hidden_states = (
                model_kwargs["encoder_outputs"].get("hidden_states") if output_hidden_states else None
            )

        # keep track of which sequences are already finished
        unfinished_sequences = input_ids.new(input_ids.shape[0]).fill_(1)
        cur_len = input_ids.shape[-1]

        logits_ids = input_ids.clone()
        this_peer_finished = False  # used by synced_gpus only
        # auto-regressive generation
        while True:

            if synced_gpus:
                # Under synced_gpus the `forward` call must continue until all gpus complete their sequence.
                # The following logic allows an early break if all peers finished generating their sequence
                this_peer_finished_flag = torch.tensor(0.0 if this_peer_finished else 1.0).to(input_ids.device)
                # send 0.0 if we finished, 1.0 otherwise
                dist.all_reduce(this_peer_finished_flag, op=dist.ReduceOp.SUM)
                # did all peers finish? the reduced sum will be 0.0 then
                if this_peer_finished_flag.item() == 0.0:
                    break

            # prepare model inputs
            model_inputs = self.prepare_inputs_for_generation(input_ids, **model_kwargs)

            # forward pass to get next token
            outputs = self(
                **model_inputs,
                return_dict=True,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                code_type=code_type,
                )

            if synced_gpus and this_peer_finished:
                cur_len = cur_len + 1
                continue  # don't waste resources running the code we don't need
            
            next_token_logits = outputs.logits[:, -1, :]
            
            # pre-process distribution
            next_token_scores = logits_processor(logits_ids, next_token_logits)

            next_token_scores = logits_warper(logits_ids, next_token_scores)

            # Store scores, attentions and hidden_states when required
            if return_dict_in_generate:
                if output_scores:
                    scores += (next_token_scores,)
                if output_attentions:
                    decoder_attentions += (
                        (outputs.decoder_attentions,) if self.config.is_encoder_decoder else (outputs.attentions,)
                    )
                    if self.config.is_encoder_decoder:
                        cross_attentions += (outputs.cross_attentions,)

                if output_hidden_states:
                    decoder_hidden_states += (
                        (outputs.decoder_hidden_states,)
                        if self.config.is_encoder_decoder
                        else (outputs.hidden_states,)
                    )

            # sample
            probs = nn.functional.softmax(next_token_scores, dim=-1)
            if probs.isnan().sum() == probs.shape[1]:
                logger.warning(f'probability of next tokens in {code_type} is NaN, do random sampling.')
                probs = torch.ones_like(probs).softmax(dim=-1)
            
            next_tokens = torch.multinomial(probs, num_samples=1).squeeze(1)

            # finished sentences should have their next token be a padding token
            if eos_token_id is not None:
                assert pad_token_id is not None, "If eos_token_id is defined, make sure that pad_token_id is defined."
                next_tokens = next_tokens * unfinished_sequences + pad_token_id * (1 - unfinished_sequences)

            # update generated ids, model inputs, and length for next step
            new_next_tokens = next_tokens.cpu().numpy() + self.model_tokenizer.label_offset
            new_next_tokens_str = self.model_tokenizer.tokenizer_dict[code_type].decode(new_next_tokens)
            if new_next_tokens_str == '':
                warnings.warn(f"Unknown token {new_next_tokens} found in {code_type},"
                            " seems the training data does not contain these tokens but the test data does."
                            " Please check your data."
                            " Will randomly pick a token here!")
                
                # randomly pick a token to fill
                random_token = '<unk>'
                not_select_list = self.data_tokenizer(f'<{code_type}> </{code_type}>', add_special_tokens=False)['input_ids']
                not_select_list = [str(x) for x in not_select_list]
                not_select_list += ['<unk>']
                all_keys = list(self.model_tokenizer.tokenizer_dict[code_type].get_vocab().keys())
                while random_token in not_select_list:
                    random_token = np.random.choice(all_keys,1)[0]
                random_token = int(random_token)

                new_next_tokens = torch.ones_like(torch.tensor(new_next_tokens)) * random_token
                new_next_tokens = new_next_tokens.to(next_tokens.device)
            else:
                new_next_tokens = new_next_tokens_str.split()
                new_next_tokens = np.array(new_next_tokens).astype(int)
                new_next_tokens = torch.tensor(new_next_tokens).to(next_tokens.device)

            input_ids = torch.cat([input_ids, new_next_tokens[:, None]], dim=-1)
            logits_ids = torch.cat([logits_ids, next_tokens[:, None]], dim=-1)

            model_kwargs = self._update_model_kwargs_for_generation(
                outputs, model_kwargs, is_encoder_decoder=self.config.is_encoder_decoder
            )
            cur_len = cur_len + 1

            # if eos_token was found in one sentence, set sentence to finished
            if eos_token_id is not None:
                unfinished_sequences = unfinished_sequences.mul((next_tokens != eos_token_id).long())

            # stop when each sentence is finished, or if we exceed the maximum length
            if unfinished_sequences.max() == 0 or stopping_criteria(input_ids, scores):
                if not synced_gpus:
                    break
                else:
                    this_peer_finished = True

        if return_dict_in_generate:
            if self.config.is_encoder_decoder:
                return SampleEncoderDecoderOutput(
                    sequences=input_ids,
                    scores=scores,
                    encoder_attentions=encoder_attentions,
                    encoder_hidden_states=encoder_hidden_states,
                    decoder_attentions=decoder_attentions,
                    cross_attentions=cross_attentions,
                    decoder_hidden_states=decoder_hidden_states,
                )
            else:
                return SampleDecoderOnlyOutput(
                    sequences=input_ids,
                    scores=scores,
                    attentions=decoder_attentions,
                    hidden_states=decoder_hidden_states,
                )
        else:
            return input_ids

    def _prepare_input_ids_for_generation(
        self, bos_token_id: Optional[int], encoder_outputs: Optional[ModelOutput]
    ) -> torch.LongTensor:
        if self.config.is_encoder_decoder and encoder_outputs is not None:
            # make dummy input_ids with value -100, as a sanity check ensuring that they won't be used for encoding
            shape = encoder_outputs.last_hidden_state.size()[:-1]
            return torch.ones(shape, dtype=torch.long, device=self.device) * -100

        # randomly initialized if no input_ids are given
        return torch.ones((1, 1), dtype=torch.long, device=self.device) * bos_token_id

    def _prepare_encoder_decoder_kwargs_for_generation(
        self, inputs_tensor: torch.Tensor, model_kwargs, model_input_name: Optional[str] = None
    ) -> Dict[str, Any]:

        # 1. get encoder & conditional prompt encoder
        promt_encoder = self.get_prompt_encoder()
        encoder = self.get_encoder()

        # 2. prepare encoder args and encoder kwargs from model kwargs
        irrelevant_prefix = ["decoder_", "cross_attn", "use_cache"]
        encoder_kwargs = {
            argument: value
            for argument, value in model_kwargs.items()
            if not any(argument.startswith(p) for p in irrelevant_prefix)
        }

        # 2.5. encode prompts if exist as the condition passed to encoder
        prompt_embeds = None
        if ('x_num' in encoder_kwargs or 'x_cat' in encoder_kwargs):
            x_num = encoder_kwargs.pop('x_num')
            x_cat = encoder_kwargs.pop('x_cat')
            if promt_encoder is not None:
                prompt_embeds = promt_encoder(x_num=x_num, x_cat=x_cat)
        encoder_kwargs['inputs_prompt_embeds'] = prompt_embeds

        # 3. make sure that encoder returns `ModelOutput`
        model_input_name = model_input_name if model_input_name is not None else self.main_input_name
        encoder_kwargs["return_dict"] = True
        encoder_kwargs[model_input_name] = inputs_tensor
        model_kwargs["encoder_outputs"]: ModelOutput = encoder(**encoder_kwargs)

        return model_kwargs

    def _prepare_decoder_input_ids_for_generation(
        self,
        batch_size: int,
        decoder_start_token_id: int = None,
        bos_token_id: int = None,
        model_kwargs: Optional[Dict[str, torch.Tensor]] = None,
        device: torch.device = None,
    ) -> torch.LongTensor:
        if model_kwargs is not None and "decoder_input_ids" in model_kwargs:
            return model_kwargs.pop("decoder_input_ids")
        else:
            decoder_start_token_id = self._get_decoder_start_token_id(decoder_start_token_id, bos_token_id)
            if device is None:
                device = self.device
            return torch.ones((batch_size, 1), dtype=torch.long, device=device) * decoder_start_token_id

    def _get_logits_processor(
        self,
        repetition_penalty: float,
        no_repeat_ngram_size: int,
        encoder_no_repeat_ngram_size: int,
        input_ids_seq_length: int,
        encoder_input_ids: torch.LongTensor,
        bad_words_ids: List[List[int]],
        min_length: int,
        max_length: int,
        eos_token_id: int,
        forced_bos_token_id: int,
        forced_eos_token_id: int,
        prefix_allowed_tokens_fn: Callable[[int, torch.Tensor], List[int]],
        num_beams: int,
        num_beam_groups: int,
        diversity_penalty: float,
        remove_invalid_values: bool,
        exponential_decay_length_penalty: Tuple,
        logits_processor: Optional[LogitsProcessorList],
        renormalize_logits: Optional[bool],
    ) -> LogitsProcessorList:
        """
        This class returns a [`LogitsProcessorList`] list object that contains all relevant [`LogitsProcessor`]
        instances used to modify the scores of the language model head.
        """
        processors = LogitsProcessorList()

        # init warp parameters
        repetition_penalty = repetition_penalty if repetition_penalty is not None else self.config.repetition_penalty
        no_repeat_ngram_size = (
            no_repeat_ngram_size if no_repeat_ngram_size is not None else self.config.no_repeat_ngram_size
        )
        encoder_no_repeat_ngram_size = (
            encoder_no_repeat_ngram_size
            if encoder_no_repeat_ngram_size is not None
            else self.config.encoder_no_repeat_ngram_size
        )
        bad_words_ids = bad_words_ids if bad_words_ids is not None else self.config.bad_words_ids
        eos_token_id = eos_token_id if eos_token_id is not None else self.config.eos_token_id
        diversity_penalty = diversity_penalty if diversity_penalty is not None else self.config.diversity_penalty
        forced_bos_token_id = (
            forced_bos_token_id if forced_bos_token_id is not None else self.config.forced_bos_token_id
        )
        forced_eos_token_id = (
            forced_eos_token_id if forced_eos_token_id is not None else self.config.forced_eos_token_id
        )
        remove_invalid_values = (
            remove_invalid_values if remove_invalid_values is not None else self.config.remove_invalid_values
        )
        exponential_decay_length_penalty = (
            exponential_decay_length_penalty
            if exponential_decay_length_penalty is not None
            else self.config.exponential_decay_length_penalty
        )
        # instantiate processors list

        # the following idea is largely copied from this PR: https://github.com/huggingface/transformers/pull/5420/files
        # all samplers can be found in `generation_utils_samplers.py`
        if diversity_penalty is not None and diversity_penalty > 0.0:
            processors.append(
                HammingDiversityLogitsProcessor(
                    diversity_penalty=diversity_penalty, num_beams=num_beams, num_beam_groups=num_beam_groups
                )
            )
        if repetition_penalty is not None and repetition_penalty != 1.0:
            processors.append(RepetitionPenaltyLogitsProcessor(penalty=repetition_penalty))
        if no_repeat_ngram_size is not None and no_repeat_ngram_size > 0:
            processors.append(NoRepeatNGramLogitsProcessor(no_repeat_ngram_size))
        if encoder_no_repeat_ngram_size is not None and encoder_no_repeat_ngram_size > 0:
            if self.config.is_encoder_decoder:
                processors.append(EncoderNoRepeatNGramLogitsProcessor(encoder_no_repeat_ngram_size, encoder_input_ids))
            else:
                raise ValueError(
                    "It's impossible to use `encoder_no_repeat_ngram_size` with decoder-only architecture"
                )
        if bad_words_ids is not None:
            processors.append(NoBadWordsLogitsProcessor(bad_words_ids, eos_token_id))
        if min_length is not None and eos_token_id is not None and min_length > 0:
            processors.append(MinLengthLogitsProcessor(min_length, eos_token_id))
        if prefix_allowed_tokens_fn is not None:
            processors.append(PrefixConstrainedLogitsProcessor(prefix_allowed_tokens_fn, num_beams // num_beam_groups))
        if forced_bos_token_id is not None:
            processors.append(ForcedBOSTokenLogitsProcessor(forced_bos_token_id))
        if forced_eos_token_id is not None:
            processors.append(ForcedEOSTokenLogitsProcessor(max_length, forced_eos_token_id))
        if remove_invalid_values is True:
            processors.append(InfNanRemoveLogitsProcessor())
        if exponential_decay_length_penalty is not None:
            processors.append(
                ExponentialDecayLengthPenalty(exponential_decay_length_penalty, eos_token_id, input_ids_seq_length)
            )
        processors = self._merge_criteria_processor_list(processors, logits_processor)
        
        # `LogitNormalization` should always be the last logit processor, when present
        if renormalize_logits is True:
            processors.append(LogitNormalization())
        return processors


from collections import defaultdict
import pdb
import time

import numpy as np
import torch
from torch import nn
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader

from pytrial.data.patient_data import SequencePatientBase, SeqPatientCollator
from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from .base import SequenceSimulationBase
from .base import InputEventEmbedding, GAN, RNN
from ..losses import GeneratorLoss, DiscriminatorLoss, DiscriminatorLossGP
from ..trainer import SeqSimGANTrainer
from ..data import SequencePatient

class BuildModel(nn.Module):
    def __init__(self,
        rnn_type,
        emb_size,
        bidirectional,
        vocab_size,
        orders,
        n_rnn_layer,
        padding_idx,
        **kwargs,
        ) -> None:
        super().__init__()
        if not isinstance(vocab_size, list): vocab_size = [vocab_size]

        # generator dim size
        # emb_size*2, gendim[0], gendim[2]: default
        # emb_size*4, gendim[0], gendim[1]: directional
        if bidirectional:
            gan_input_emb_size = 2*emb_size
            gen_dims=[gan_input_emb_size, 2*emb_size]
        else:
            gan_input_emb_size = emb_size
            gen_dims=[2*emb_size, 2*emb_size]

        # discriminator dim size
        dis_dims=[2*emb_size]

        if bidirectional:
            gen_dims = [g*2 for g in gen_dims]

        self.gan_module = GAN(
            emb_size=gan_input_emb_size,
            total_vocab_size=sum(vocab_size),
            gen_dims=gen_dims,
            dis_dims=dis_dims,
        )

        self.rnn_module = RNN(
            rnn_type=rnn_type,
            emb_size=emb_size,
            num_layer=n_rnn_layer,
            bidirectional=bidirectional,
        )

        self.vocab_size = vocab_size
        self.orders = orders

        self.embeddings = InputEventEmbedding(orders=orders, vocab_size=vocab_size, emb_size=emb_size, padding_idx=padding_idx)

    def forward(self, inputs, n=1):
        '''
        inputs is a dict
        {
            'v': {'eventA':[],'eventB':[],...,},
            'x': tensor(),
            'y': {'eventA:[], 'eventB':[],...,} # optional
        }
        '''
        embs = self.embeddings(inputs) # [num_visit, num_event, emb_size]
        embs = torch.sum(embs, 1) # [num_visit, emb_size]
        embs = self.rnn_module(embs) # [num_visit, emb_size] or [num_visit, emb_size*2] (bidirectional)
        last_visit_emb = embs[-1] # [emb_size]

        if len(last_visit_emb.shape) == 1:
            last_visit_emb = last_visit_emb.unsqueeze(0)

        if n > 1:
            # generate more than one visits
            last_visit_emb = last_visit_emb.expand(n, -1)

        # infer generator
        z_random = torch.randn(last_visit_emb.size()).to(last_visit_emb.device)
        x_fake = self.gan_module.infer_generator(z_random, last_visit_emb) # 1, vocab_size
        
        # infer discriminator for real records
        if 'y' in inputs:
            # infer discriminator for fake records
            y_fake = self.gan_module.infer_discriminator(x_fake)
            target = self._create_multilabel_target(inputs)
            target = target.to(x_fake.device)
            y_real = self.gan_module.infer_discriminator(target.float())
            return {'x_fake':x_fake, 'y_real':y_real, 'y_fake':y_fake, 'y':target}
        else:
            return {'x_fake':x_fake}
    
    def infer_discriminator(self, x):
        return self.gan_module.infer_discriminator(x)

    def _create_multilabel_target(self, inputs):
        # for one patient one visit only `inputs`
        target = inputs['y']
        target_mo_list = []
        for i,o in enumerate(self.orders):
            target_mo = torch.zeros(self.vocab_size[i], dtype=torch.long)
            target_mo[target[o]] = 1
            target_mo_list.append(target_mo)
        targets = torch.cat(target_mo_list)
        if len(targets.shape) == 1: targets = targets.unsqueeze(0)
        return targets


class RNNGAN(SequenceSimulationBase):
    '''
    Implement an RNN based GAN model for longitudinal patient records simulation. The GAN part was proposed by Choi et al. [1]_.
    
    Parameters
    ----------
    vocab_size: list[int]
        A list of vocabulary size for different types of events, e.g., for diagnosis, procedure, medication.

    order: list[str]
        The order of event types in each visits, e.g., ``['diag', 'prod', 'med']``.
        Visit = [diag_events, prod_events, med_events], each event is a list of codes.

    max_visit: int
        The maximum number of visits for input event codes.

    emb_size: int
        Embedding size for encoding input event codes.

    n_rnn_layer: int
        Number of RNN layers for encoding historical events.

    rnn_type: str
        Pick RNN types in ['rnn','lstm','gru']

    bidirectional: bool
        If True, it encodes historical events in bi-directional manner.
    
    padding_idx: int(default=None)
        Set the padding index for input events embedding. If set None, then no
        padding index will be specified.

    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    weigth_decay: float
        Regularization strength for l2 norm; must be a positive float.
        Smaller values specify weaker regularization.
    
    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    num_worker: int
        Number of workers used to do dataloading during training.

    device: str
        The model device.

    Notes
    -----
    .. [1] Choi, E., et al. (2017, November). Generating multi-label discrete patient records using generative adversarial networks. In ML4HC (pp. 286-305). PMLR.
    '''
    def __init__(self,
        vocab_size,
        order,
        max_visit=20,
        emb_size=64,
        n_rnn_layer=2,
        rnn_type='lstm',
        bidirectional=False,
        padding_idx=None,
        learning_rate=1e-4,
        weight_decay=1e-4,
        batch_size=64,
        epochs=10,
        num_worker=0,
        device='cuda:0',
        experiment_id='trial_simulation.sequence.rnn_gan',
        ):
        super().__init__(experiment_id)
        self.config = {
            'vocab_size':vocab_size,
            'max_visit':max_visit,
            'emb_size':emb_size,
            'n_rnn_layer':n_rnn_layer,
            'rnn_type':rnn_type,
            'bidirectional':bidirectional,
            'padding_idx':padding_idx,
            'device':device,
            'learning_rate':learning_rate,
            'batch_size':batch_size,
            'weight_decay':weight_decay,
            'epochs':epochs,
            'num_worker':num_worker,
            'orders':order,
            }
        self.config['total_vocab_size'] = sum(vocab_size)
        self.device = device
        self._build_model()
    
    def fit(self, train_data):
        '''
        Train model with sequential patient records.

        Parameters
        ----------
        train_data: SequencePatientBase
            A `SequencePatientBase` contains patient records where 'v' corresponds to 
            visit sequence of different events.
        '''
        self._input_data_check(train_data)
        self._fit_model(train_data)

    def predict(self, test_data, n=None, n_per_sample=None, return_tensor=True):
        '''
        Generate synthetic records based on input real patient seq data.

        Parameters
        ----------
        test_data: SequencePatientBase
            A `SequencePatientBase` contains patient records where 'v' corresponds to 
            visit sequence of different events.

        n: int
            How many samples in total will be generated.

        n_per_sample: int
            How many samples generated based on each indivudals.
        
        return_tensor: bool
            If `True`, return output generated records in tensor format (n, n_visit, n_event), good for later predictive modeling.
            If `False, return records in `SequencePatient` format.
        '''
        if n is not None: assert isinstance(n, int), 'Input `n` should be integer.'
        if n_per_sample is not None: assert isinstance(n_per_sample, int), 'Input `n_per_sample` should be integer.'
        n, n_per_sample = self._compute_n_per_sample(len(test_data), n, n_per_sample)
        test_dataloader = self.get_test_dataloader(test_data)
        outputs = self._predict_on_dataloader(test_dataloader, n, n_per_sample)
        if not return_tensor:
            outputs = self._translate_sparse_visits_to_dense(outputs)
        else:
            # pad all to same shape
            outputs = self._pad_multiple_tensor_visits(outputs)
        return outputs

    def save_model(self, output_dir):
        '''
        Save the learned simulation model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir

        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model}, output_dir=output_dir)

    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.
            
            - If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            - If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        self.model = state_dict['model']
    
    def get_train_dataloader(self, train_data):
        dataloader = DataLoader(train_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=True,
            collate_fn=SeqPatientCollator(
                config={
                    'visit_mode':train_data.metadata['visit']['mode'],
                    'label_mode':train_data.metadata['label']['mode'],
                    }
                ),
            )
        return dataloader

    def get_test_dataloader(self, train_data):
        dataloader = DataLoader(train_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=False,
            shuffle=False,
            collate_fn=SeqPatientCollator(
                config={
                    'visit_mode':train_data.metadata['visit']['mode'],
                    'label_mode':train_data.metadata['label']['mode'],
                    }
                ),
            )
        return dataloader

    def _build_model(self):
        self.model = BuildModel(
            rnn_type=self.config['rnn_type'],
            emb_size=self.config['emb_size'],
            max_visit=self.config['max_visit'],
            n_rnn_layer=self.config['n_rnn_layer'],
            bidirectional=self.config['bidirectional'],
            vocab_size=self.config['vocab_size'],
            orders=self.config['orders'],
            padding_idx=self.config['padding_idx'],
            )
        self.model.to(self.device)
    
    def _build_loss_model(self):
        # update discriminator two times v.s. generator one time
        return [DiscriminatorLoss(self.model), GeneratorLoss(self.model)]

    def _fit_model(self, train_data):
        train_dataloader = self.get_train_dataloader(train_data)
        loss_models = self._build_loss_model()
        train_objectives = [(train_dataloader, loss_model) for loss_model in loss_models]
        trainer = SeqSimGANTrainer(
            model=self,
            train_objectives=train_objectives
            )
        trainer.train(**self.config)

    @torch.no_grad()
    def _predict_on_dataloader(self, test_dataloader, n, n_per_sample):
        data_iterator = iter(test_dataloader)
        total_number = 0
        fake_visit_list = []
        while total_number < n:
            try:
                data = next(data_iterator)
            except:
                data_iterator = iter(test_dataloader)
                data = next(data_iterator)
            
            for idx, _ in enumerate(data['x']):
                num_visit = self._get_num_visit(data, idx)
                if num_visit < 2: # deal with more than one visit only
                    continue
                fake_visits = []
                for vdx in range(1, num_visit):
                    inputs = self._prepare_input(data, idx, vdx)
                    if 'y' in inputs: inputs.pop('y')
                    x_fake = self.model(inputs, n=n_per_sample)['x_fake']
                    fake_visits.append(x_fake.cpu().numpy())
                
                fake_visits = np.stack(fake_visits, 1) # n_sample, n_visit, total_vocab_size
                fake_visits[fake_visits>0.5]=1
                fake_visits[fake_visits<=0.5]=0

                outputs = self._prepare_input(data, idx, 1)
                first_visit = self._translate_dense_visits_to_sparse(outputs['v'])
                first_visit = np.tile(first_visit[None],(len(fake_visits),1,1))
                fake_visits = np.concatenate([first_visit, fake_visits], 1)
                fake_visit_list.append(fake_visits) # add one synthetic record
                total_number += len(fake_visits)
                if total_number >= n: break

        return fake_visit_list

    def _translate_sparse_visits_to_dense(self, visits):
        def _map_func(x):
            res = np.where(x > 0)[0].tolist()
            return [0] if len(res) == 0 else res # pad if nothing happened

        outputs = defaultdict(list)
        for batchv in visits:
            voc_offset = 0
            for i, o in enumerate(self.config['orders']):
                voc_size = self.config['vocab_size'][i]
                visit = batchv[...,voc_offset:voc_offset+voc_size] # 10, 7, 5
                for visit_ in visit:
                    res = list(map(_map_func, visit_))
                    outputs[o].append(res)
                voc_offset += voc_size

        n_total = len(outputs[o])
        sample_list = []        
        for i in range(n_total):
            sample = []
            for numv in range(len(outputs[o][i])):
                visit = []
                for o in self.config['orders']:
                    visit.append(outputs[o][i][numv])
                sample.append(visit)
            sample_list.append(sample)

        # create seqpatient data
        return SequencePatient(
            data={'v':sample_list},
            metadata={
                'visit':{'mode':'dense','order':self.config['orders']},
                }
            )

    def _translate_dense_visits_to_sparse(self, visits):
        total_vocab_size = sum(self.config['vocab_size'])
        num_visits = len(visits[self.config['orders'][0]])
        outputs = np.zeros((num_visits, total_vocab_size))

        for i, o in enumerate(self.config['orders']):
            for j in range(num_visits):
                raw = visits[o][j]
                if isinstance(raw, torch.Tensor): raw = raw.detach().cpu().numpy()
                if i > 0: 
                    voc_size = sum(self.config['vocab_size'][:i-1])
                    if isinstance(raw, list):
                        raw = [r + voc_size for r in raw]
                    else:
                        raw += voc_size
                outputs[j, raw] = 1
        return outputs

    def _pad_multiple_tensor_visits(self, visits):
        new_list = []
        for v in visits:
            new_list.extend([torch.tensor(x).squeeze(0) for x in np.array_split(v, len(v))])
        return pad_sequence(new_list, batch_first=True)
    
    def _input_data_check(self, inputs):
        assert isinstance(inputs, SequencePatientBase), f'`trial_simulation.sequence` models require input training data in `SequencePatientBase`, find {type(inputs)} instead.'

    


import pdb    
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import Categorical

#######
### Loss Functions Components
#######
# ordered_score = (batch, numMC, M)
def compute_log_prob(rank_samples, score, num_MC, K):
	ordered_score = torch.gather(score.unsqueeze(1).repeat(1,num_MC,1), 2, rank_samples).to(score.device)
	first_k = ordered_score[:,:,:K]
	permuted_indices = torch.argsort(torch.rand(*first_k.shape), dim=-1).to(score.device)
	permuted_first_k = torch.gather(first_k, -1, permuted_indices)
	random_top_k_ordered_score = torch.cat((permuted_first_k, ordered_score[:,:,K:]), dim=-1)
	denominators = torch.flip(torch.cumsum(torch.flip(random_top_k_ordered_score, [-1]), -1), [-1])
	ranking_prob = torch.prod((random_top_k_ordered_score / denominators)[:,:,:K], -1)	
	prob = math.factorial(K) * ranking_prob
	return torch.log(prob)

def compute_prob(rank_samples, score, num_MC, K):
	ordered_score = torch.gather(score.unsqueeze(1).repeat(1,num_MC,1), 2, rank_samples).to(score.device)
	first_k = ordered_score[:,:,:K]
	permuted_indices = torch.argsort(torch.rand(*first_k.shape), dim=-1).to(score.device)
	permuted_first_k = torch.gather(first_k, -1, permuted_indices)
	random_top_k_ordered_score = torch.cat((permuted_first_k, ordered_score[:,:,K:]), dim=-1)
	denominators = torch.flip(torch.cumsum(torch.flip(random_top_k_ordered_score, [-1]), -1), [-1])
	ranking_prob = torch.prod((random_top_k_ordered_score / denominators)[:,:,:K], -1)	
	prob = math.factorial(K) * ranking_prob
	return prob

def ranking_loss(rank_samples, relevance, num_MC, K):
	relevance = relevance/torch.sum(relevance, 1, keepdim=True)
	relevance[relevance != relevance] = 0 # Handle division by zero and set to zero
	relevance = torch.gather(relevance.unsqueeze(1).repeat(1,num_MC,1), 2, rank_samples).to(relevance.device)
	return torch.sum(relevance[:,:,:K], 2).to(relevance.device) - torch.sum(relevance[:,:,K:], 2).to(relevance.device)

def fairness_loss(rank_samples, eth_list, relevance, num_MC, K):
	relevance = torch.gather(relevance.unsqueeze(1).repeat(1,num_MC,1), 2, rank_samples).to(eth_list.device)[:,:,:K]
	relevance = relevance/torch.sum(relevance, 2, keepdim=True) # Normalize the relevance into weights
	relevance[relevance != relevance] = 1 / K # Handle division by zero and set to even weighting
	eth_list = torch.gather(eth_list.unsqueeze(1).repeat(1,num_MC,1,1), 2, rank_samples.unsqueeze(-1).repeat(1,1,1,eth_list.size(-1)))[:,:,:K,:]
	eth_list = eth_list * relevance.unsqueeze(-1) # Weight the distributions
	delta_f = torch.sum(eth_list, dim=2) # Sum across the chosen K
	reward = Categorical(probs = delta_f).entropy()
	return reward

#######
### Complete Loss Functions
#######

# score = bs * M
# relevance = bs * M
# eth_list = bs * M * num_cat
# rank_samples = bs * MC * M
def loss_function(score, relevance, eth_list, lam, K):	
		num_MC_samples = 25
		M = len(score[0])
		score = F.softmax(score, dim=1)

		rank_samples = torch.stack([torch.multinomial(score, num_samples=M, replacement=False).to(score.device) for _ in range(num_MC_samples)], axis=1)

		importance_prob_values = compute_log_prob(rank_samples, score, num_MC_samples, K)
		delta_values = ranking_loss(rank_samples, relevance, num_MC_samples, K)
		delta_f_values = fairness_loss(rank_samples, eth_list, relevance, num_MC_samples, K)
		rewards = delta_values + lam*delta_f_values
		loss = -1 * torch.sum(importance_prob_values * rewards, 1)
		return loss/num_MC_samples #, torch.mean(rewards), torch.mean(delta_values), torch.mean(delta_f_values)

def loss_function_enrollment(score, relevance, K):	
		num_MC_samples = 25
		M = len(score[0])
		score = F.softmax(score, dim=1)

		rank_samples = torch.stack([torch.multinomial(score, num_samples=M, replacement=False).to(score.device) for _ in range(num_MC_samples)], axis=1)

		importance_prob_values = compute_log_prob(rank_samples, score, num_MC_samples, K)
		delta_values = ranking_loss(rank_samples, relevance, num_MC_samples, K)
		rewards = delta_values
		loss = -1 * torch.sum(importance_prob_values * rewards, 1)
		return loss/num_MC_samples #, torch.mean(rewards), torch.mean(delta_values)

def loss_function_fairness(score, relevance, eth_list, K):
		num_MC_samples = 25
		M = len(score[0])
		score = F.softmax(score, dim=1)

		rank_samples = torch.stack([torch.multinomial(score,num_samples=M, replacement=False).to(score.device) for _ in range(num_MC_samples)], axis=1)
		
		importance_prob_values = compute_log_prob(rank_samples, score, num_MC_samples, K)
		delta_f_values = fairness_loss(rank_samples, eth_list, relevance, num_MC_samples, K)

		loss = -1 * torch.sum(importance_prob_values * (delta_f_values), 1)
		return loss/num_MC_samples

class PolicyGradientLossEnrollment(nn.Module):
    def __init__(self, model, K) -> None:
        super().__init__()
        self.model = model
        self.K = K
        self.loss = loss_function_enrollment

    def forward(self, inputs):
        scores = self.model(inputs)
        loss = self.loss(scores, inputs['label'], self.K)
        return {'loss_value':loss.mean()}
    
class PolicyGradientLossFairness(nn.Module):
    def __init__(self, model, K) -> None:
        super().__init__()
        self.model = model
        self.K = K
        self.loss = loss_function_fairness

    def forward(self, inputs):
        scores = self.model(inputs)
        loss = self.loss(scores, inputs['label'], inputs['eth_label'], self.K)
        return {'loss_value':loss.mean()}
    
class PolicyGradientLossCombined(nn.Module):
    def __init__(self, model, K, lam) -> None:
        super().__init__()
        self.model = model
        self.K = K
        self.lam = lam
        self.loss = loss_function

    def forward(self, inputs):
        scores = self.model(inputs)
        loss = self.loss(scores, inputs['label'], inputs['eth_label'], self.lam, self.K)
        return {'loss_value':loss.mean()}


'''
Topic discovery for clinical trial data.
'''
import pdb
import copy

import pandas as pd
import numpy as np
from torch import nn
from sklearn.cluster import KMeans

import warnings
from trial2vec import Trial2Vec
from trial2vec import download_embedding
warnings.filterwarnings('ignore')

from .data_utils import icdcode_text_2_lst_of_lst, list_text_2_list, smiles_txt_to_lst
from .data_utils import flatten_stacked_list
from .data_utils import split_protocol
from .data_structure import Trial, Topic, SequenceTrial

class TrialTopicDiscovery:
    '''
    Input a csv of clinical trial data, output a dataset consisting of multiple sequences of 
    clinical trials. Each sequence is a list of clinical trials under this topic ordered with their timestamps.
    '''
    def __init__(self,
        num_topics=50,
        random_seed=42,
        ) -> None:
        super().__init__()
        self.num_topics = num_topics
        self.random_seed = random_seed
        self.km = None
        # trial2vec = Trial2Vec()
        # self.trial2vec.from_pretrained()
        self.trial2vec = download_embedding()
    
    def fit(self, trial_data):
        '''
        Fit the topic discovery model.

        Parameters
        ----------
        trial_data : pd.DataFrame
            A dataframe of clinical trial data.
        '''
        _ = self.topic_classification(trial_data)
    
    def transform(self, trial_data):
        '''
        Transform a list of dataframes into a `SequenceTrial` object.

        Parameters
        ----------
        trial_data : pd.DataFrame
            A dataframe of clinical trial data.

        Returns
        -------
        seqtrial : SequenceTrial
            A `SequenceTrial` object.
        '''
        trial_data.fillna('[PAD]', inplace=True)
        datalist = self.topic_classification(trial_data)
        seqtrial = self.create_sequence_trial(datalist)
        return seqtrial

    def fit_transform(self, trial_data):
        '''
        Fit the topic discovery model and transform the trial data into a dataset of sequences of trials.

        Parameters
        ----------
        trial_data : pd.DataFrame
            A dataframe of clinical trial data.

        Returns
        -------
        seqtrial : SequenceTrial
            A dataset consisting of multiple topics of clinical trials. Each topic is a list of clinical trials under this topic ordered with their timestamps.
        '''
        # at present we just use the trial2vec embeddings for topic discovery by clustering
        _ = self.fit(trial_data)

        # create a dataset of topics
        seqtrial = self.transform(trial_data)
        return seqtrial

    def topic_classification(self, trial_data):
        '''
        Create a list of dataframes, each of which is a dataframe of clinical trial data for a given topic id.

        Can be replaced by other topic classification methods. At present, clustering is used.
        '''
        datalist = self._split_by_clustering(trial_data)
        return datalist

    def create_sequence_trial(self, datalist):
        '''
        Create a `SequenceTrial` object from a list of dataframes.

        Parameters
        ----------
        data_list : List[[str, pd.DataFrame]]
            A list of dataframes, each of which is a dataframe of clinical trial data for a given topic id.

        Returns
        -------
        seqtrial : SequenceTrial
            A `SequenceTrial` object.
        '''
        topics = []
        for topic_id, df in datalist:
            df_list = self._preprocess_df(df)
            topic = self._build_topic(topic_id, df_list)
            topics.append(topic)
        seqtrial = SequenceTrial(topics)
        return seqtrial

    def add_trials(self, seqtrial, trial_data):
        '''
        Add new trials to a given `TrialSequence` object.
        Will classify trials to each topic and add them to the corresponding topic/timestamp.

        Parameters
        ----------
        seqtrial : TrialSequence
            A `TrialSequence` object.

        trial_data : pd.DataFrame
            A dataframe of clinical trial data.
        '''
        new_seqtrial = copy.deepcopy(seqtrial)
        data_list = self.topic_classification(trial_data)
        topic_list = []
        for topic_id, df in data_list:
            df_list = self._preprocess_df(df)
            topic = self._build_topic(topic_id, df_list)
            topic_list.append(topic)
        new_seqtrial.add(topic_list)
        return new_seqtrial

    def _split_by_clustering(self, df):
        '''
        Split the dataframe trials by clustering based on Trial2Vec pretrained embeddings.

        Parameters
        ----------
        df : pd.DataFrame
            A dataframe of clinical trial data.

        Returns
        -------
        df_list : List[[str, pd.DataFrame]]
            A list of dataframes, each of which is a dataframe of clinical trial data for a given year.
        '''
        # get the embeddings
        emblist = [self.trial2vec[nctid] for nctid in df['nctid']]

        # cluster the embeddings
        if self.km is None:
            self.km = KMeans(n_clusters=self.num_topics, random_state=self.random_seed)
            self.km.fit(emblist)
            topic_ids = self.km.predict(emblist)
        else:
            topic_ids = self.km.predict(emblist)

        # split the dataframe by topic ids
        df_list = []
        for topic_id in range(self.num_topics):
            df_list.append([str(topic_id), df[topic_ids == topic_id]])
        return df_list

    def _split_by_year(self, df):
        '''
        Split the dataframe by year.

        Parameters
        ----------
        df : pd.DataFrame
            A dataframe of clinical trial data.

        Returns
        -------
        df_list : List[pd.DataFrame]
            A list of dataframes, each of which is a dataframe of clinical trial data for a given year.
        '''
        df_list = []
        df = df.sort_values(by='year')
        for year, df_year in df.groupby('year'):
            df_list.append([str(year), df_year])
        return df_list

    def _preprocess_df(self, df):
        ecs = df['criteria'].apply(split_protocol)
        df['inclusion_criteria'] = ecs.apply(lambda x: x[0])
        df['exclusion_criteria'] = ecs.apply(lambda x: x[1] if len(x) > 1 else ['[PAD]'])
        df.drop(columns=['criteria'], inplace=True)
        df['icdcodes'] = df['icdcodes'].apply(icdcode_text_2_lst_of_lst)
        df['diseases'] = df['diseases'].apply(list_text_2_list)
        df['drugs'] = df['drugs'].apply(list_text_2_list)
        df['smiless'] = df['smiless'].apply(smiles_txt_to_lst)

        # get year and sort and split by year
        df['study_first_submitted_date'] = pd.to_datetime(df['study_first_submitted_date'])
        df['year'] = df['study_first_submitted_date'].dt.year
        df_list = self._split_by_year(df)
        return df_list
    
    def _build_topic(self, topic_id, df_list):
        '''
        Build a topic object from a list of dataframes.

        Parameters
        ----------
        topic_id : str
            The topic id.

        df_list : List[pd.DataFrame]
            A list of dataframes, each of which is a dataframe of clinical trial data for a given year.

        Returns
        -------
        topic : Topic
            A topic object.
        '''
        trials = []
        for year, df_year in df_list:
            trial_year = []
            for _, row in df_year.iterrows():
                trial = self._build_trial(row)
                trial_year.append(trial)
            trials.append(trial_year)
        topic = Topic(topic_id, trials)
        return topic
    
    def _build_trial(self, row):
        '''
        Build a trial object from a row of a dataframe.

        Parameters
        ----------
        row : pd.Series
            A row of a dataframe.

        Returns
        -------
        trial : Trial
            A trial object.
        '''
        row = row.to_dict()
        trial = Trial(
            nctid=row.pop('nctid'),
            title=row.pop('title'),
            label=row.pop('label'),
            status=row.pop('status'),
            year=row.pop('year'),
            **row,
            )
        return trial


'''
Provide the basic vocabular instance.
'''

class Vocab(object):
    def __init__(self):
        self.idx2word = {}
        self.word2idx = {}
    
    def __len__(self):
        return len(self.idx2word.keys())
    
    @property
    def words(self):
        '''All the words in the vocab.

        Returns
        -------
        words: list[str]
        '''
        return list(self.word2idx.keys())
    
    @property
    def vocab(self):
        '''The vocabulary where key is the index and value is the word.

        Returns
        -------
        vocab: dict[int, str]
        '''
        return self.idx2word

    def add_sentence(self, sentence):
        '''
        Add a list of words to the vocabulary. If one word is in the vocab, then ignore it.
        Otherwise, add it to the vocab.
        
        Parameters
        ----------
        sentence : list[str]
            A list of words.
        '''
        if isinstance(sentence, list):
            if len(sentence) == 0:
                return    
            for word in sentence:
                self._add_word(word)
        else:
            self._add_word(sentence)
    
    def _add_word(self, word):
        if word not in self.word2idx:
            self.idx2word[len(self.word2idx)] = word
            self.word2idx[word] = len(self.word2idx)


import sdmetrics


def _select_metrics(metrics):
    '''
    Select multiple metrics at once.

    Parameters
    ----------

    metrics (list[str]):
        List of metric names to apply or None:
            metrics = [
                'BNLogLikelihood',
                'LogisticDetection',
                'SVCDetection',
                'BinaryDecisionTreeClassifier',
                'BinaryAdaBoostClassifier',
                'BinaryLogisticRegression',
                'BinaryMLPClassifier',
                'MulticlassDecisionTreeClassifier',
                'MulticlassMLPClassifier',
                'LinearRegression',
                'MLPRegressor',
                'GMLogLikelihood',
                'CSTest',
                'KSComplement',
                'StatisticSimilarity',
                'BoundaryAdherence',
                'MissingValueSimilarity',
                'CategoryCoverage',
                'TVComplement',
                'CategoricalCAP',
                'CategoricalZeroCAP',
                'CategoricalGeneralizedCAP',
                'CategoricalNB',
                'CategoricalKNN',
                'CategoricalRF',
                'CategoricalSVM',
                'CategoricalEnsemble',
                'NumericalLR',
                'NumericalMLP',
                'NumericalSVR',
                'NumericalRadiusNearestNeighbor',
                'ContinuousKLDivergence',
                'DiscreteKLDivergence',
                'ContingencySimilarity',
                'CorrelationSimilarity',
            ]

    Returns
    -------
    final_metrics: sdmetrics.single_table.SingleTableMetric
    '''
    metric_classes = sdmetrics.single_table.SingleTableMetric.get_subclasses()
    if metrics is None:
        metric_classes = {
            'KSComplement': metric_classes['KSComplement'],
            'CSTest': metric_classes['CSTest'],
        }
        return metric_classes

    final_metrics = {}
    for metric in metrics:
        if isinstance(metric, str):
            try:
                final_metrics[metric] = metric_classes[metric]
            except KeyError:
                raise ValueError(f'Unknown metric: {metric}')

    return final_metrics

def evaluate(synthetic_data, real_data, metrics=None, aggregate=True):
    '''
    Apply multiple metrics at once.

    Parameters
    ----------
    synthetic_data (dict[str, pandas.DataFrame] or pandas.DataFrame):
        Map of names and tables of synthesized data. When evaluating a single table,
        a single ``pandas.DataFrame`` can be passed alone.

    real_data (dict[str, pandas.DataFrame] or pandas.DataFrame):
        Map of names and tables of real data. When evaluating a single table,
        a single ``pandas.DataFrame`` can be passed alone.
        If metadata is None, this parameter must be a dataframe.

    metrics (list[str]):
        List of metric names to apply:
            'BNLogLikelihood'<class 'sdmetrics.single_table.bayesian_network.BNLogLikelihood'>,
            'LogisticDetection': <class 'sdmetrics.single_table.detection.sklearn.LogisticDetection'>, 
            'SVCDetection': <class 'sdmetrics.single_table.detection.sklearn.SVCDetection'>, 
            'BinaryDecisionTreeClassifier': <class 'sdmetrics.single_table.efficacy.binary.BinaryDecisionTreeClassifier'>, 
            'BinaryAdaBoostClassifier': <class 'sdmetrics.single_table.efficacy.binary.BinaryAdaBoostClassifier'>, 
            'BinaryLogisticRegression': <class 'sdmetrics.single_table.efficacy.binary.BinaryLogisticRegression'>, 
            'BinaryMLPClassifier': <class 'sdmetrics.single_table.efficacy.binary.BinaryMLPClassifier'>, 
            'MulticlassDecisionTreeClassifier': <class 'sdmetrics.single_table.efficacy.multiclass.MulticlassDecisionTreeClassifier'>, 
            'MulticlassMLPClassifier': <class 'sdmetrics.single_table.efficacy.multiclass.MulticlassMLPClassifier'>, 
            'LinearRegression': <class 'sdmetrics.single_table.efficacy.regression.LinearRegression'>, 
            'MLPRegressor': <class 'sdmetrics.single_table.efficacy.regression.MLPRegressor'>, 
            'GMLogLikelihood': <class 'sdmetrics.single_table.gaussian_mixture.GMLogLikelihood'>, 
            'CSTest': <class 'sdmetrics.single_table.multi_single_column.CSTest'>, 
            'KSComplement': <class 'sdmetrics.single_table.multi_single_column.KSComplement'>, 
            'StatisticSimilarity': <class 'sdmetrics.single_table.multi_single_column.StatisticSimilarity'>, 
            'BoundaryAdherence': <class 'sdmetrics.single_table.multi_single_column.BoundaryAdherence'>, 
            'MissingValueSimilarity': <class 'sdmetrics.single_table.multi_single_column.MissingValueSimilarity'>, 
            'CategoryCoverage': <class 'sdmetrics.single_table.multi_single_column.CategoryCoverage'>, 
            'TVComplement': <class 'sdmetrics.single_table.multi_single_column.TVComplement'>, 
            'CategoricalCAP': <class 'sdmetrics.single_table.privacy.cap.CategoricalCAP'>, 
            'CategoricalZeroCAP': <class 'sdmetrics.single_table.privacy.cap.CategoricalZeroCAP'>, 
            'CategoricalGeneralizedCAP': <class 'sdmetrics.single_table.privacy.cap.CategoricalGeneralizedCAP'>, 
            'CategoricalNB': <class 'sdmetrics.single_table.privacy.categorical_sklearn.CategoricalNB'>, 
            'CategoricalKNN': <class 'sdmetrics.single_table.privacy.categorical_sklearn.CategoricalKNN'>, 
            'CategoricalRF': <class 'sdmetrics.single_table.privacy.categorical_sklearn.CategoricalRF'>, 
            'CategoricalSVM': <class 'sdmetrics.single_table.privacy.categorical_sklearn.CategoricalSVM'>, 
            'CategoricalEnsemble': <class 'sdmetrics.single_table.privacy.ensemble.CategoricalEnsemble'>, 
            'NumericalLR': <class 'sdmetrics.single_table.privacy.numerical_sklearn.NumericalLR'>, 
            'NumericalMLP': <class 'sdmetrics.single_table.privacy.numerical_sklearn.NumericalMLP'>, 
            'NumericalSVR': <class 'sdmetrics.single_table.privacy.numerical_sklearn.NumericalSVR'>, 
            'NumericalRadiusNearestNeighbor': <class 'sdmetrics.single_table.privacy.radius_nearest_neighbor.NumericalRadiusNearestNeighbor'>, 
            'ContinuousKLDivergence': <class 'sdmetrics.single_table.multi_column_pairs.ContinuousKLDivergence'>, 
            'DiscreteKLDivergence': <class 'sdmetrics.single_table.multi_column_pairs.DiscreteKLDivergence'>, 
            'ContingencySimilarity': <class 'sdmetrics.single_table.multi_column_pairs.ContingencySimilarity'>, 
            'CorrelationSimilarity': <class 'sdmetrics.single_table.multi_column_pairs.CorrelationSimilarity'>,
    
    aggregate (bool):
        If ``get_report`` is ``False``, whether to compute the mean of all the normalized
        scores to return a single float value or return a ``dict`` containing the score
        that each metric obtained. Defaults to ``True``.
    
    Returns
    -------
    scores: float or sdmetrics.MetricsReport
    '''
    metrics = _select_metrics(metrics)
    scores = sdmetrics.compute_metrics(metrics, real_data, synthetic_data)

    if aggregate:
        return scores.normalized_score.mean()

    return scores



from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
import pdb

import torch
from torch import nn
from torch.utils.data import DataLoader

from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from .base import SiteSelectionBase
from .data import TrialSiteModalities, SiteSelectionModalitiesCollator
from .losses import PolicyGradientLossEnrollment, PolicyGradientLossCombined
from .trainer import SiteSelectTrainer

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

#######
### FRAMM Implementations
#######

class DiagnosisEncoder(nn.Module):
    '''
    Build a site's diagnosis history encoder.
    '''
    def __init__(self, 
        dx_dim, 
        lstm_dim, 
        embedding_dim, 
        num_layers):
        super(DiagnosisEncoder, self).__init__()
        
        self.dx_dim = dx_dim
        self.lstm_dim = lstm_dim
        self.embedding_dim = embedding_dim
        self.num_layers = num_layers
        self.biLSTM = nn.LSTM(dx_dim, lstm_dim, num_layers, batch_first=True, bidirectional=True)
        self.fc = nn.Linear(2*lstm_dim, embedding_dim)
        self.firstInput = nn.Parameter(torch.rand(1,1,1,dx_dim, dtype=torch.float))

    def forward(self, input, lengths, num_inv):
        bs = input.size(0)
        input = torch.cat((self.firstInput.repeat(bs, num_inv, 1, 1), input), -2)
        input = torch.reshape(input, (bs * num_inv, -1, self.dx_dim))
        lengths = torch.reshape(lengths, (bs * num_inv,))
        packed_input = pack_padded_sequence(input, lengths+1, batch_first=True, enforce_sorted=False)
        packed_output, _ = self.biLSTM(packed_input)
        out, _ = pad_packed_sequence(packed_output, batch_first=True)
        combined_out = torch.cat((out[:, 0, self.lstm_dim:], out[range(len(out)), lengths, :self.lstm_dim]), 1)
        combined_out = torch.reshape(combined_out, (bs, num_inv, 2 * self.lstm_dim))
        return(self.fc(combined_out))

class PrescriptionEncoder(nn.Module):
    '''
    Build a site's prescription history encoder.
    '''
    def __init__(self, 
        rx_dim, 
        lstm_dim, 
        embedding_dim, 
        num_layers):
        super(PrescriptionEncoder, self).__init__()
        
        self.rx_dim = rx_dim
        self.lstm_dim = lstm_dim
        self.embedding_dim = embedding_dim
        self.num_layers = num_layers
        self.biLSTM = nn.LSTM(rx_dim, lstm_dim, num_layers, batch_first=True, bidirectional=True)
        self.fc = nn.Linear(2*lstm_dim, embedding_dim)
        self.firstInput = nn.Parameter(torch.rand(1,1,1,rx_dim, dtype=torch.float))

    def forward(self, input, lengths, num_inv):
        bs = input.size(0)
        input = torch.cat((self.firstInput.repeat(bs, num_inv, 1, 1), input), -2)
        input = torch.reshape(input, (bs * num_inv, -1, self.rx_dim))
        lengths = torch.reshape(lengths, (bs * num_inv,))
        packed_input = pack_padded_sequence(input, lengths+1, batch_first=True, enforce_sorted=False)
        packed_output, _ = self.biLSTM(packed_input)
        out, _ = pad_packed_sequence(packed_output, batch_first=True)
        combined_out = torch.cat((out[:, 0, self.lstm_dim:], out[range(len(out)), lengths, :self.lstm_dim]), 1)
        combined_out = torch.reshape(combined_out, (bs, num_inv, 2 * self.lstm_dim))
        return(self.fc(combined_out))

class PastTrialEncoder(nn.Module):
    '''
    Build a site's trial history encoder.
    '''
    def __init__(self, 
        trial_dim, 
        lstm_dim, 
        embedding_dim, 
        num_layers):
        super(PastTrialEncoder, self).__init__()
        
        self.trial_dim = trial_dim
        self.lstm_dim = lstm_dim
        self.embedding_dim = embedding_dim
        self.num_layers = num_layers
        self.biLSTM = nn.LSTM(trial_dim+1, lstm_dim, num_layers, batch_first=True, bidirectional=True)
        self.fc = nn.Linear(2*lstm_dim, embedding_dim)
        self.firstInput = nn.Parameter(torch.rand(1,1,1,trial_dim+1, dtype=torch.float))

    def forward(self, input, lengths, num_inv):
        bs = input.size(0)
        input = torch.cat((self.firstInput.repeat(bs, num_inv, 1, 1), input), -2)
        input = torch.reshape(input, (bs * num_inv, -1, self.trial_dim+1))
        lengths = torch.reshape(lengths, (bs * num_inv,))
        packed_input = pack_padded_sequence(input, lengths+1, batch_first=True, enforce_sorted=False)
        packed_output, _ = self.biLSTM(packed_input)
        out, _ = pad_packed_sequence(packed_output, batch_first=True)
        combined_out = torch.cat((out[:, 0, self.lstm_dim:], out[range(len(out)), lengths, :self.lstm_dim]), -1)
        combined_out = torch.reshape(combined_out, (bs, num_inv, 2 * self.lstm_dim))
        return(self.fc(combined_out))
  
class AttentionEncoder(nn.Module):
    '''
    Build a site's representation encoder which combines each of the possibly missing modalities.
    '''
    def __init__(self, 
        embed_dim, 
        n_modalities,
        n_heads=4):
        super(AttentionEncoder, self).__init__()
        self.embed_dim = embed_dim
        self.n_modalities = n_modalities
        self.attention = nn.MultiheadAttention(embed_dim, n_heads)

    def forward(self, query, values, attention_mask):
        values = torch.stack(values, dim=2)
        bs = values.size(0)
        num_inv = values.size(1)
        query = torch.reshape(query, (bs * num_inv, 1, self.embed_dim)).transpose(0,1)
        values = torch.reshape(values, (bs * num_inv, self.n_modalities, self.embed_dim)).transpose(0,1)
        attention_mask = torch.reshape(attention_mask, (bs * num_inv, self.n_modalities)).bool()
        embeddings, _ = self.attention(query, values, values, attention_mask, need_weights=False)
        embeddings = torch.reshape(embeddings.transpose(0,1), (bs, num_inv, self.embed_dim))
        return embeddings
    
class ModalityDropoutEncoder(nn.Module):
    '''
    Build a site's representation encoder which combines each of the possibly missing modalities.
    '''
    def __init__(self, 
        embed_dim, 
        n_modalities):
        super(ModalityDropoutEncoder, self).__init__()
        self.embed_dim = embed_dim
        self.n_modalities = n_modalities
        self.encoder = nn.Linear(n_modalities*embed_dim, embed_dim)

    def forward(self, query, values, attention_mask):
        for i in range(self.n_modalities):
            values[i] = values[i] * attention_mask[:,:,i]
        values = torch.cat(values, dim=-1)
        embeddings = self.encoder(values)
        return embeddings
    
class TransformerScoringNetwork(nn.Module):
    '''
    Generate score representations using a Transformer encoder layer
    '''
    def __init__(self, embedding_dim, n_heads, hidden_dim):
        encoder_layer = nn.TransformerEncoderLayer(2*embedding_dim, n_heads, hidden_dim)
        self.encoder = nn.TransformerEncoder(encoder_layer, 2)
    
    def forward(self, inv_site_representation):
        return self.encoder(inv_site_representation.transpose(0,1)).transpose(0,1)
  
class BuildModel(nn.Module):
    def __init__(self, 
        trial_dim, 
        static_dim, 
        dx_dim, 
        rx_dim, 
        lstm_dim, 
        embedding_dim, 
        num_layers, 
        hidden_dim, 
        n_heads,
        missing_type='MCAT',
        scoring_type='Transformer'):
        super().__init__()
        
        self.static_encoder = nn.Linear(static_dim, embedding_dim)
        self.diagnosis_encoder = DiagnosisEncoder(dx_dim, lstm_dim, embedding_dim, num_layers)
        self.prescription_encoder = PrescriptionEncoder(rx_dim, lstm_dim, embedding_dim, num_layers)
        self.history_encoder = PastTrialEncoder(trial_dim, lstm_dim, embedding_dim, num_layers)
        self.trial_encoder = nn.Linear(trial_dim, embedding_dim)
        self.stat_fc = nn.Linear(embedding_dim, embedding_dim)
        self.dx_fc = nn.Linear(embedding_dim, embedding_dim)
        self.rx_fc = nn.Linear(embedding_dim, embedding_dim)
        self.hist_fc = nn.Linear(embedding_dim, embedding_dim)
        self.trial_fc = nn.Linear(embedding_dim, embedding_dim)
        if missing_type == 'MCAT':
            self.missing_modality_encoder = AttentionEncoder(embedding_dim, 4, n_heads)
        elif missing_type == 'MD':
            self.missing_modality_encoder = ModalityDropoutEncoder(embedding_dim, 4)
        else:
            raise ValueError(f'Invalid missing_type: {missing_type}')
        if scoring_type == 'Transformer':
            self.score_encoder = TransformerScoringNetwork(embedding_dim, n_heads, hidden_dim)
        elif scoring_type == 'Fully Connected':
            self.score_encoder = nn.Linear(2*embedding_dim, 2*embedding_dim)
        else:
            raise ValueError(f'Invalid scoring_type: {scoring_type}')
        self.fc = nn.Linear(2*embedding_dim, hidden_dim)
        self.output = nn.Linear(hidden_dim, 1)
        
    def forward(self, inputs):
        trial = inputs['trial']
        investigators = inputs['inv_static']
        dx = inputs['dx']
        dx_lens = inputs['dx_len']
        rx = inputs['rx']
        rx_lens = inputs['rx_len']
        past_trials = inputs['enroll_hist']
        past_trials_lengths = inputs['enroll_hist_len']
        inv_mask = inputs['inv_mask']
        num_inv = investigators.size(1)
        # Trial is (bs, trial_dim)
        # All other inputs are (bs, M, *) where * is either a sequence or single input
        investigator_encoding = self.stat_fc(torch.relu(self.static_encoder(investigators)))
        dx_encoding = self.dx_fc(torch.relu(self.diagnosis_encoder(dx, dx_lens, num_inv)))
        rx_encoding = self.rx_fc(torch.relu(self.prescription_encoder(rx, rx_lens, num_inv)))
        history_encoding = self.hist_fc(torch.relu(self.history_encoder(past_trials, past_trials_lengths, num_inv)))
        trial_encoding = self.trial_fc(torch.relu(self.trial_encoder(trial)))
        trial_encoding = trial_encoding.repeat(1, num_inv, 1)
        inv_representation = self.missing_modality_encoder(trial_encoding, [investigator_encoding, dx_encoding, rx_encoding, history_encoding], inv_mask)
        inv_site_representation = torch.cat((inv_representation, trial_encoding), dim=-1)
        network_input = self.score_encoder(inv_site_representation)
        score = self.output(torch.relu(self.fc(network_input)))
        return score.squeeze(-1)

class FRAMM(SiteSelectionBase):
    '''
    Implement FRAMM for selecting clinical trial sites based on possibly missing multi-model site features [1]_.
    
    Parameters
    ----------
    
    trial_dim: list[int]
        Size of the trial representation

    static_dim: int
        Size of the static site features modality

    dx_dim: int
        Size of the diagnosis code vocabulary

    rx_dim: int
        Size of the prescription code vocabulary

    lstm_dim: int
        Size of the lstm embedding

    embedding_dim: int
        Size of all of the modality and other intermediate embeddings

    num_layers: int
        Number of lstm layers
    
    hidden_dim: int
        Size of intermediate representation within output layer

    n_heads: int
        Number of heads for attention encoder
        
    missing_type: string
        Type of missing data mechanism to use
        
    scoring_type: string
        Type of scoring network to use
    
    Notes
    -----
    .. [1] Theodorou, B., Glass, L., Xiao, C., & Sun, J. (2023). FRAMM: Fair Ranking with Missing Modalities for Clinical Trial Site Selection. arXiv preprint arXiv:2305.19407.
        
    '''
    def __init__(self, 
        trial_dim=211, 
        static_dim=124, 
        dx_dim=157, 
        rx_dim=79, 
        lstm_dim=64, 
        embedding_dim=64, 
        num_layers=2, 
        hidden_dim=32, 
        n_heads=4,
        missing_type='MCAT',
        scoring_type='Transformer',
        enrollment_only=True,
        K=10,
        lam=1,
        learning_rate=1e-4,
        weight_decay=1e-4,
        batch_size=64,
        epochs=10,
        num_worker=0,
        device='cuda:0',
        experiment_id='test',
        ) -> None:
        super().__init__(experiment_id)
        self.config = {
            'trial_dim':trial_dim,
            'static_dim':static_dim,
            'dx_dim':dx_dim,
            'rx_dim':rx_dim,
            'lstm_dim':lstm_dim,
            'embedding_dim':embedding_dim,
            'num_layers':num_layers,
            'hidden_dim':hidden_dim,
            'n_heads':n_heads,
            'missing_type':missing_type,
            'scoring_type':scoring_type,
            'enrollment_only':enrollment_only,
            'K':K,
            'lambda':lam,
            'learning_rate':learning_rate,
            'weight_decay':weight_decay,
            'batch_size':batch_size,
            'epochs':epochs,
            'num_worker':num_worker,
            'device':device,
            'experiment_id':experiment_id,
            }
        self.device = device
        self._build_model()

    def fit(self, train_data):
        '''
        Train model with historical trial-site enrollments.

        Parameters
        ----------
        train_data: TrialSiteModalities
            A `TrialSiteModalities` contains trials, sites, and enrollments.
        '''
        self._input_data_check(train_data)
        self._fit_model(train_data)

    def predict(self, test_data):
        self._input_data_check(test_data)
        test_dataloader = self.get_train_dataloader(test_data)
        selections = []
        for batch in test_dataloader:
            batch = self._prepare_input(batch)
            scores = self.model(batch)
            batch_selections = torch.argsort(scores, dim=1, descending=True)[:, :self.config['K']].tolist()
            for selection in batch_selections:
                selections.append(selection)
                
        return selections

    def save_model(self, output_dir):
        '''
        Save the learned patient-match model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir
        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model}, output_dir=output_dir)

    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.
            If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        self.model = state_dict['model']
    
    def get_train_dataloader(self, train_data):
        dataloader = DataLoader(train_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=True,
            collate_fn=SiteSelectionModalitiesCollator(
                config={
                    'visit_mode':train_data.sites.metadata['visit']['mode'],
                    'trial_mode':train_data.sites.metadata['hist']['mode'],
                    'has_demographics':'true' if train_data.sites.get_label() is not None else 'false'
                    }
                ),
            )
        return dataloader

    def _build_model(self):
        self.model = BuildModel(
            trial_dim=self.config['trial_dim'],
            static_dim=self.config['static_dim'],
            dx_dim=self.config['dx_dim'],
            rx_dim=self.config['rx_dim'],
            lstm_dim=self.config['lstm_dim'],
            embedding_dim=self.config['embedding_dim'],
            num_layers=self.config['num_layers'],
            hidden_dim=self.config['hidden_dim'],
            n_heads=self.config['n_heads'],
            missing_type=self.config['missing_type'],
            scoring_type=self.config['scoring_type']
            )
    
    def _build_loss_model(self):
        if self.config['enrollment_only']:
            return PolicyGradientLossEnrollment(self.model, self.config['K'])
        else:
            return PolicyGradientLossCombined(self.model, self.config['K'], self.config['lambda'])     

    def _fit_model(self, train_data):
        train_dataloader = self.get_train_dataloader(train_data)
        loss_model = self._build_loss_model()
        train_objectives = [(train_dataloader, loss_model)]
        trainer = SiteSelectTrainer(
            model=self,
            train_objectives=train_objectives
            )
        trainer.train(**self.config)
        
    def _prepare_input(self, data):
        '''
        Prepare inputs to model.

        Parameters
        ----------
        data: dict[list]
            A batch of trials with their corresponding sites.
        '''
        inputs = {
            'trial': data['trial'].to(self.device),
            'label': data['label'].to(self.device),
            'eth_label': None if data['eth_label'] is None else data['eth_label'].to(self.device),
            'inv_static': data['inv_static'].to(self.device),
            'dx': data['dx'].to(self.device),
            'dx_len': data['dx_len'].to('cpu'),
            'rx': data['rx'].to(self.device),
            'rx_len': data['rx_len'].to('cpu'),
            'enroll_hist': data['enroll_hist'].to(self.device),
            'enroll_hist_len': data['enroll_hist_len'].to('cpu'),
            'inv_mask': data['inv_mask'].to(self.device),
            }

        return inputs

    def _input_data_check(self, inputs):
        assert isinstance(inputs, TrialSiteModalities), f'`site_selection` models with multi-modality support require input training data in `TrialSiteModalities`, find {type(inputs)} instead.'


'''
User interface to use promptEHR models.
'''
import os
import pdb
import json
import math
import glob
import random
import copy
from collections import defaultdict
import warnings

import dill
import torch
from torch import nn
from torch.utils.data.dataloader import DataLoader
from transformers import TrainingArguments
import numpy as np
from tqdm import tqdm

from .dataset import MimicDataset, MimicDataCollator
from .modeling_config import EHRBartConfig, DataTokenizer, ModelTokenizer
from .trainer import PromptEHRTrainer
from .evaluator import Evaluator
from .model import BartForEHRSimulation
from . import constants

class PromptEHR(nn.Module):
    '''
    Initialize a PromptEHR model to leverage language models to simulate sequential patient EHR data.

    Parameters:
    -----------
    code_type: list[str]
        A list of code types that the model will learn and generate.
        For example, `code_type=['diag','prod','med']`.

    token_dict: dict[list]
        A dictionary of new tokens (code events, e.g., ICD code) that the model needs to learn and generate.

    n_num_feature: int (default=None)
        Number of numerical patient baseline features. Notice that it assumes that the input
        baseline features are `ALWAYS` numerical feature first. That is to say,
        the input baseline feature = [num1, num2, .., num_n, cat1, cat2,...].
        If not specified, the model will never include baseline features
        for conditional generation!

    cat_cardinalities: list[int]
        The number of categories for each categorical patient baseline features.
        The input baseline feature = [num1, num2, .., num_n, cat1, cat2,...].

    epoch: int
        Num training epochs in total.

    batch_size: int
        Training batch size.

    eval_batch_size: int
        Evaluation batch size.
    
    eval_step: int
        How many steps of updates then try to evaluate the trained models.
    
    learning_rate: float
        Training learning rate.
    
    weight_decay: float
        Training weight decay.
    
    num_worker: int
        Numer of dataloading paralleled processes.
    
    output_dir: str
        Training logs output to this folder.

    device: str or list[int]
        Should be str like `cuda:0` or `cpu`, otherwise should be a list GPU ids.
    '''
    sample_config = {
        'num_beams': 1, # >1: beam_sample; =1: sample_gen
        'no_repeat_ngram_size': 1,
        'do_sample': True,
        'num_return_sequences': 1,
        'code_type': 'diagnosis',
        'top_k': 1,
        'temperature': 1.0,
        'max_length': 6,
    }
    def __init__(self,
        code_type=None,
        n_num_feature=None,
        cat_cardinalities=None,
        epoch=50,
        batch_size=16,
        eval_batch_size=16,
        eval_step=1000,
        learning_rate=5e-5,
        weight_decay=1e-4,
        num_worker=8,
        output_dir='./promptEHR_logs',
        device='cuda:0',
        seed=123,
        ) -> None:
        super().__init__()
        self.data_tokenizer = DataTokenizer.from_pretrained('facebook/bart-base')

        # will extend vocab after pass training data
        if code_type is not None:
            self.data_tokenizer.update_special_token_config(code_types=code_type)
        self.model_tokenizer = None
        self.config = {
            'code_type': code_type,
            'n_num_feature':n_num_feature,
            'cat_cardinalities':cat_cardinalities,
            'epoch':epoch,
            'batch_size':batch_size,
            'eval_batch_size':eval_batch_size,
            'eval_step':eval_step,
            'learning_rate':learning_rate,
            'weight_decay':weight_decay,
        }
        self.device = device
        if isinstance(device, list):
            self._set_visible_device(device=device)
        self.training_args = TrainingArguments(
            per_device_train_batch_size=batch_size, # no. of patients in each batch, default to be 1
            per_device_eval_batch_size=eval_batch_size,
            gradient_accumulation_steps=1,
            learning_rate=learning_rate,
            weight_decay=weight_decay,
            output_dir=output_dir,
            num_train_epochs=epoch,
            save_steps=eval_step,
            eval_steps=eval_step,
            warmup_ratio=0.06,
            max_grad_norm=0.5,
            save_total_limit=5,
            logging_steps=eval_step,
            dataloader_num_workers=num_worker, # debug
            dataloader_pin_memory=True,
            evaluation_strategy='steps',
            metric_for_best_model=f'eval_ppl_{code_type[0]}' if code_type is not None else None,
            greater_is_better=False, # NLL is the less the better
            eval_accumulation_steps=10,
            load_best_model_at_end=True,
            logging_dir=output_dir,      # directory for storing logs
            overwrite_output_dir=True,
            seed=seed,
            no_cuda=True if self.device == 'cpu' else False, # if set CPU
            )

        # avoid dead clock when taking multiple workers for dataloaders
        os.environ['TOKENIZERS_PARALLELISM'] = 'false'

        self.model = None

    def fit(self, train_data, val_data=None):
        '''
        Fit PromptEHR model on the input training EHR data.

        Parameters
        ----------
        train_data: SequencePatient
            A `SequencePatient` contains patient records where 'v' corresponds to 
            visit sequence of different events.

        val_data: dict
            A `SequencePatient` contains patient records where 'v' corresponds to 
            visit sequence of different events.
        '''
        # create tokenizers based on the input data
        self._create_tokenizers(train_data)

        # can only build model after fit
        self._build_model()

        # start training
        self._fit(train_data=train_data,val_data=val_data)
    
    def predict(self, test_data, n_per_sample=None, n=None, sample_config=None, verbose=None):
        '''
        Generate synthetic records based on input real patient seq data.

        Parameters
        ----------
        test_data: SequencePatient
            A `SequencePatient` contains patient records where 'v' corresponds to 
            visit sequence of different events.

        n: int
            How many samples in total will be generated.

        n_per_sample: int
            How many samples generated based on each indivudals.

        sample_config: dict
            Configuration for sampling synthetic records, key parameters:
            'num_beams': Number of beams in beam search, if set `1` then beam search is deactivated;
            'top_k': Sampling from top k candidates.
            'temperature': temperature to make sampling distribution flater or skewer.
        
        verbose: bool
            If print the progress bar or not.

        Returns
        -------
        Synthetic patient records in `SequencePatient` format.
        '''
        if n is not None: assert isinstance(n, int), 'Input `n` should be integer.'
        if n_per_sample is not None: assert isinstance(n_per_sample, int), 'Input `n_per_sample` should be integer.'
        assert (not n_per_sample is None) or (not n is None), 'Either `n` or `n_per_sample` should be provided to generate.'
        assert isinstance(self.model, BartForEHRSimulation), 'Model not found! Please fit the model or load the model from pretrained checkpoint first.'

        n, n_per_sample = self._compute_n_per_sample(len(test_data), n, n_per_sample)

        if sample_config is not None:
            self.sample_config.update(sample_config)
            print('### Sampling Config ###')
            print(self.sample_config)

        # get test data loader
        test_dataloader = self._get_test_dataloader(test_data)

        # make generation
        outputs = self._predict_on_dataloader(test_dataloader, n, n_per_sample, verbose=verbose)

        # formulate outputs to standard sequencepatient data
        # need 'visit', 'order', 'feature', 'n_num_feature', 'cat_cardinalties'
        visits, features, labels = [], [], []
        for output in outputs:
            code_types = [c for c in self.config['code_type'] if c in output]
            num_visit = len(output[code_types[0]])
            visit, feature = [], []
            for n in range(num_visit):
                visit_ = [output[code][n] for code in code_types]
                visit.append(visit_)
            visits.append(visit)
            if 'x_num' in output:
                feature.extend(output['x_num'])
            if 'x_cat' in output:
                feature.extend(output['x_cat'])
            if len(feature) > 0:
                features.append(feature)
            if 'y' in output: labels.append(output['y'])
        
        if len(features) > 0:
            features = np.stack(features, 0)
        else:
            features = None

        return_res = {
            'visit':visits, 
            'feature':features, 
            'order':self.config['code_type'],
            'n_num_feature':self.config['n_num_feature'],
            'cat_cardinalties':self.config['cat_cardinalities'],
            'y':labels,
            'voc': test_data.metadata['voc'],
        }
        return return_res

    def save_model(self, output_dir):
        '''
        Save the learned simulation model to the disk.

        Parameters
        ----------
        output_dir: str
            The dir to save the learned model.
        '''
        make_dir_if_not_exist(output_dir)
        self._save_config(config=self.config, output_dir=output_dir)
        self._save_checkpoint(output_dir=output_dir)
        print('Save the trained model to:', output_dir)
    
    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        data_tokenizer_file, model_tokenizer_file = check_tokenizer_file(checkpoint)

        # load config
        self.config = self._load_config(config_filename)

        # load data tokenizer and model tokenizer
        self._load_tokenizer(data_tokenizer_file, model_tokenizer_file)

        # load configuration
        self.configuration = EHRBartConfig(self.data_tokenizer, self.model_tokenizer, n_num_feature=self.config['n_num_feature'], cat_cardinalities=self.config['cat_cardinalities'])
        self.configuration.from_pretrained(checkpoint)

        # build model
        self._build_model()

        # load checkpoint
        state_dict = torch.load(checkpoint_filename, map_location='cpu')
        self.load_state_dict(state_dict, strict=True)
        print('Load the pre-trained model from:', checkpoint)

    def evaluate(self, test_data):
        '''
        Evaluate the trained PromptEHR model on the input data, will test the perplexity
        for each type of codes.
        
        Parameters
        ----------
        test_data: PatientSequence
            Standard sequential patient records in `PatientSequence` format.
        '''
        self.model.eval()
        self.eval()

        collator = MimicDataCollator(
            self.data_tokenizer,
            code_types=self.config['code_type'],
            n_num_feature=self.config['n_num_feature'],
            mode='test', 
            drop_feature=False
            )

        evaluator = Evaluator(
            self.model,
            test_data,
            collator,
            device='cpu' if self.device == 'cpu' else 'cuda:0',
        )

        code_types = self.config['code_type']
        ppl_types = ['tpl','spl']
        for code_type in code_types:
            for ppl_type in ppl_types:
                ppl = evaluator.evaluate(code_type, ppl_type, eval_batch_size=self.config['eval_batch_size'])
                print(f'code: {code_type}, ppl_type: {ppl_type}, value: {ppl}')

    def from_pretrained(self, input_dir='./simulation/pretrained_promptEHR'):
        '''
        Load pretrained PromptEHR model and make patient EHRs generation.
        Pretrained model was learned from MIMIC-III patient sequence data.
        '''
        if input_dir is None:
            input_dir = './simulation/pretrained_promptEHR'
        
        if not os.path.exists(input_dir):
            os.makedirs(input_dir)            
            url = constants.PRETRAINED_MODEL_URL
            download_pretrained(url, input_dir)
            print(f'Download pretrained PromptEHR model, save to {input_dir}.')
        
        print('Load pretrained PromptEHR model from', input_dir)
        self.load_model(input_dir)
    
    def update_config(self, config):
        '''
        Update the configuration of the model.

        Parameters
        ----------
        config: dict
            The configuration of the model.
            Refer to the `config` in `__init__` for more details.
        '''
        self.config.update(config)
        
        # update training args
        train_args = copy.deepcopy(config)
        for k, v in config.items():
            if k in constants.config_to_train_args:
                train_args[constants.config_to_train_args[k]] = v
                train_args.pop(k)
        
        for k,v in train_args.items():
            if hasattr(self.training_args, k):
                setattr(self.training_args, k, v)
        
        # important when you train the model with different datasets
        code_type = self.config['code_type']
        self.training_args.metric_for_best_model = \
            f'eval_ppl_{code_type[0]}' if code_type is not None else None,

        print('### Model Config ###')
        print(self.config)

        print('### Training Args ###')
        print(self.training_args)

    def _save_config(self, config, output_dir=None):        
        temp_path = os.path.join(output_dir, 'promptehr_config.json')
        with open(temp_path, 'w', encoding='utf-8') as f:
            f.write(
                json.dumps(config, indent=4)
            )

        # save the data tokenizer and model tokenizer of the model
        temp_path = os.path.join(output_dir, 'data_tokenizer.pkl')
        with open(temp_path, 'wb') as f:
            dill.dump(self.data_tokenizer, f)

        temp_path = os.path.join(output_dir, 'model_tokenizer.pkl')
        with open(temp_path, 'wb') as f:
            dill.dump(self.model_tokenizer, f)

        # save configuration
        self.configuration.save_pretrained(output_dir)

    def _load_tokenizer(self, data_tokenizer_file, model_tokenizer_file):
        with open(data_tokenizer_file, 'rb') as f:
            self.data_tokenizer = dill.load(f)
        self.data_tokenizer._in_target_context_manager = False # fix bugs when upgrade transformers to 4.23

        with open(model_tokenizer_file, 'rb') as f:
            self.model_tokenizer = dill.load(f)

    def _load_config(self, filename):
        with open(filename, 'r') as f:
            config = json.load(f)
        return config

    def _get_test_dataloader(self, dataset):
        def _seq_patient_to_promptehr(samples):
            post_samples = []
            for sample in samples:
                post_sample = {}
                visit = sample['v']
                post_sample.update(visit)

                if ('x' in sample) and (self.config['n_num_feature'] is not None):
                    if not isinstance(sample['x'], list): 
                        sample['x'] = sample['x'].tolist()
                    post_sample['x_num'] = torch.tensor(sample['x'][:self.config['n_num_feature']])
                    post_sample['x_cat'] = torch.tensor(sample['x'][self.config['n_num_feature']:], dtype=int)

                if 'y' in sample:
                    post_sample['y'] = sample['y']

                post_samples.append(post_sample)
            return post_samples

        dataloader = DataLoader(dataset,
                batch_size=1, # one patient once
                drop_last=False,
                num_workers=0,
                pin_memory=False,
                shuffle=False,
                collate_fn=_seq_patient_to_promptehr,
                )
        return dataloader

    def _save_checkpoint(self,
                        epoch_id=0,
                        is_best=False,
                        output_dir=None,
                        filename='checkpoint.pth.tar'):

        if epoch_id < 1:
            filepath = os.path.join(output_dir, 'latest.' + filename)
        elif is_best:
            filepath = os.path.join(output_dir, 'best.' + filename)
        else:
            filepath = os.path.join(output_dir, str(epoch_id) + '.' + filename)
        
        # save statedict
        state_dict = self.state_dict()
        torch.save(state_dict, filepath)

    def _fit(self, train_data, val_data):
        mimic_train_collator = MimicDataCollator(self.data_tokenizer, 
            code_types=self.config['code_type'],
            n_num_feature=self.config['n_num_feature'],
            max_train_batch_size=self.config['batch_size'], mode='train')

        mimic_val_collator = MimicDataCollator(self.data_tokenizer, 
            code_types=self.config['code_type'],
            n_num_feature=self.config['n_num_feature'],
            mode='val')

        trainer = PromptEHRTrainer(
            model=self.model,
            args=self.training_args,
            train_dataset=train_data,
            data_collator=mimic_train_collator,
            eval_dataset=val_data,
            val_data_collator=mimic_val_collator,
            )
        trainer.train()
    
    def _create_tokenizers(self, train_data):
        # update data_tokenizer first
        def _collate_fn(inputs):
            outputs = defaultdict(list)
            for input in inputs:
                visit = input['v']
                for k,v in visit.items():
                    code_list = sum(v,[])
                    code_list = [k+'_'+str(c) for c in list(set(code_list))]
                    outputs[k].extend(code_list)
            return outputs
        dataloader = DataLoader(train_data, collate_fn=_collate_fn, batch_size=512, shuffle=False)
        for batch in dataloader:
            for k,v in batch.items():
                unq_codes = list(set(v))
                self.data_tokenizer.add_token_to_code_vocab(unq_codes, k)
        
        self.model_tokenizer = ModelTokenizer(self.data_tokenizer)
        self.configuration = EHRBartConfig(self.data_tokenizer, self.model_tokenizer, n_num_feature=self.config['n_num_feature'], cat_cardinalities=self.config['cat_cardinalities'])
        self.data_tokenizer.update_special_token_config(code_types=self.config['code_type'])

    def _build_model(self):
        self.model = BartForEHRSimulation(self.configuration, self.model_tokenizer, self.data_tokenizer)

        # check if cuda is available using torch
        if not torch.cuda.is_available():
            warnings.warn('No GPU found, using CPU instead.')
            self.device = 'cpu'

        if isinstance(self.device, list): 
            self._set_visible_device(self.device)
            self.model.cuda()
        elif 'cuda' in self.device: 
            self.model.cuda()
        else:
            # on cpu
            self._set_visible_device([])
            self.model.cpu()

    def _set_visible_device(self, device):
        if len(device) > 0:
            os.environ['CUDA_VISIBLE_DEVICES'] = ','.join([str(d) for d in device])
        else:
            os.environ['CUDA_VISIBLE_DEVICES'] = ''

    def _compute_n_per_sample(self, n_test_sample, n=None, n_per_sample=None):
        if n_per_sample is not None:
            n_total = n_test_sample*n_per_sample
            if n is not None:
                n_total = min(n_total, n)
            return n_total, n_per_sample
        else:
            return n, math.ceil(n / n_test_sample)

    def _get_num_visit(self, data, idx):
        visits = data['v']
        num_visit_list = []
        for k,v in visits.items():
            num_visit_list.append(len(v[idx]))
        
        num_visit_uq = list(set(num_visit_list))
        assert len(num_visit_uq) == 1, f'Find mismatch in the number of visit events {num_visit_list}, please check the input data {visits}.'
        return num_visit_uq[0]

    def _predict_on_dataloader(self, dataloader, n, n_per_sample, verbose=None):
        total_number = 0
        data_iterator = iter(dataloader)

        if verbose:
            pbar = tqdm(total=n)

        new_record_list = []
        while total_number < n:
            try:
                data = next(data_iterator)
            except:
                data_iterator = iter(dataloader)
                data = next(data_iterator)            
            data = data[0] # batch size is 1 when doing generation

            # to device
            device = 'cpu' if self.device == 'cpu' else 'cuda:0'
            if 'x_num' in data: data['x_num'] = data['x_num'].to(device)
            if 'x_cat' in data: data['x_cat'] = data['x_cat'].to(device)                
            
            inputs = self._prepare_input_for_generation(data) 

            # start generation
            for _ in range(n_per_sample):
                new_record = self._generation_loop(data, inputs)
                if 'x_cat' in data:
                    new_record.update({
                        'x_cat':data['x_cat'].cpu().numpy().tolist(),
                    })
                    
                if 'x_num' in data:
                    new_record.update({
                        'x_num':data['x_num'].cpu().numpy().tolist(),
                    })

                # add more features to new_record
                for k,v in data.items():
                    if k not in new_record:
                        new_record[k] = v
                new_record_list.append(new_record)
            
            total_number += n_per_sample
            if verbose:
                pbar.update(n_per_sample)
                    
        if verbose:
            pbar.close()
        return new_record_list

    def _prepare_input_for_generation(self, data):        
        def _process_span(span, code):
            return [code+'_'+str(s) for s in span]
        
        def _to_device(x, device):
            for k,v in x.items():
                x[k] = v.to(device)
            return x

        tokenizer = self.data_tokenizer
        code_type = [k for k in data.keys() if k in self.config['code_type']]
        num_visit = len(data[code_type[0]])
        
        # init codes
        init_code = random.sample(data[code_type[0]][0], 1)
        init_code_str = _process_span(init_code, code_type[0])
        init_codes = tokenizer(init_code_str, return_tensors='pt', add_special_tokens=False)
        bos = torch.tensor([tokenizer.bos_token_id])
        code_prompt_idx = tokenizer.encode(tokenizer.special_token_dict[code_type[0]], add_special_tokens=False, return_tensors='pt')
        init_input_ids = torch.cat([bos[:,None],code_prompt_idx[:,0,None],init_codes['input_ids']], dim=-1)
        init_input_ids = _to_device({'input_ids':init_input_ids}, self.model.device)['input_ids']
        input_ids = init_input_ids.clone()
        return {'input_ids':input_ids, 'init_input_ids':init_input_ids, 'num_visit':num_visit, 'init_code':init_code}

    def _generation_loop(self, data, inputs):
        new_record = defaultdict(list)
        tokenizer = self.data_tokenizer
        special_token_dict = self.data_tokenizer.special_token_dict
        sample_gen_kwargs = self.sample_config.copy()

        input_ids_list = []
        num_visit_code_list = []
        first_code_flag = True

        input_ids = inputs['input_ids']
        for visit in range(inputs['num_visit']):
            this_visit_ids_list = []
            for code in self.config['code_type']:
                target_list = data[code][visit]
                sample_gen_kwargs['code_type'] = code
                num_code = len(target_list)
                if num_code > 20:
                    num_code = min(num_code, 20)
                    target_list = np.random.choice(target_list, num_code, replace=False).tolist()

                # random select part of codes from target list
                target_ar = np.array(target_list)
                sub_code = target_ar[np.random.binomial(1, 0.5, num_code).astype(bool)]
                code_prompt_idx = [special_token_dict[code][0]] + sub_code.tolist() + [special_token_dict[code][1]]
                code_prompt_idx = tokenizer.encode(code_prompt_idx, add_special_tokens=False, return_tensors='pt')
                code_prompt_idx = code_prompt_idx.to(self.model.device)

                if num_code == 0:
                    if first_code_flag:
                        new_next_tokens = code_prompt_idx[:,-1,None]
                        first_code_flag = False
                    else:
                        new_next_tokens = code_prompt_idx

                    this_visit_ids_list.append(new_next_tokens)
                    input_ids = torch.cat([input_ids, new_next_tokens], dim=-1)
                    new_record[code].append([])
                
                else:
                    sample_gen_kwargs['max_length'] = num_code+2

                    # do conditional generation
                    if 'x_cat' in data:
                        sample_gen_kwargs['x_cat'] = data['x_cat']
                    if 'x_num' in data:
                        sample_gen_kwargs['x_num'] = data['x_num']

                    new_next_tokens = self.model.generate(input_ids, **sample_gen_kwargs)

                    # randomly pick / rm sub code overlap
                    new_next_tokens = new_next_tokens[:,1:-1]
                    new_next_tokens = np.setdiff1d(new_next_tokens[0].cpu(), code_prompt_idx[0].cpu())
                    
                    try:
                        if num_code-len(sub_code) > len(new_next_tokens):
                            new_sub_idxs = np.unique(np.random.choice(np.arange(len(new_next_tokens)), num_code-len(sub_code), replace=True))
                        else:
                            new_sub_idxs = np.unique(np.random.choice(np.arange(len(new_next_tokens)), num_code-len(sub_code), replace=False))
                    except:
                        pdb.set_trace()
                        pass
                    new_next_tokens = torch.tensor(new_next_tokens[None, new_sub_idxs]).to(code_prompt_idx.device)

                    # append to the synthetic record dict
                    code_str_list = tokenizer.batch_decode(new_next_tokens)[0]

                    # remove special tokens ahead of original code event
                    # e.g., `diag_384` -> `384`
                    code_str_list = code_str_list.replace(code+'_','')
                    code_str_list = code_str_list.split()
                    code_str_list = [int(c) for c in code_str_list+sub_code.tolist()]
                    new_record[code].append(list(set(code_str_list)))

                    if first_code_flag:
                        new_next_tokens = torch.cat([new_next_tokens, code_prompt_idx[:,1:]], dim=-1)
                        first_code_flag = False
                    else:
                        # cover by modality prompt
                        new_next_tokens = torch.cat([code_prompt_idx[:,:-1], new_next_tokens, code_prompt_idx[:,-1,None]], dim=-1)

                    if visit > 1:
                        # check input length
                        cur_len = input_ids.shape[1] + new_next_tokens.shape[1]
                        while cur_len >= tokenizer.model_max_length:
                            print(f'{cur_len} reach model max length {tokenizer.model_max_length}, do cut.')
                            input_ids_list = input_ids_list[1:]
                            num_visit_code_list = num_visit_code_list[1:]
                            input_ids = torch.cat(input_ids_list,dim=-1)
                            cur_len = input_ids.shape[1] + new_next_tokens.shape[1]

                    # concat
                    this_visit_ids_list.append(new_next_tokens)
                    input_ids = torch.cat([input_ids, new_next_tokens], dim=-1)

            # after one visit, add eos token id
            eos = torch.tensor([tokenizer.eos_token_id], device=self.model.device)
            input_ids = torch.cat([input_ids, eos[:,None]], dim=-1)
            this_visit_ids = torch.cat(this_visit_ids_list, dim=-1)
            this_visit_ids = torch.cat([this_visit_ids, eos[:,None]], dim=-1)
            if visit == 0: this_visit_ids = torch.cat([inputs['init_input_ids'], this_visit_ids], dim=-1)
            num_visit_code_list.append(this_visit_ids.shape[-1])
            input_ids_list.append(this_visit_ids)

        # add init code
        new_record[self.config['code_type'][0]][0] += inputs['init_code']
        return new_record


def download_pretrained(url, output_dir):
    import wget
    import zipfile
    filename = wget.download(url=url, out=output_dir)
    zipf = zipfile.ZipFile(filename, 'r')
    zipf.extractall(output_dir)
    zipf.close()

def make_dir_if_not_exist(path):
    if not os.path.exists(path):
        os.makedirs(path)

def check_checkpoint_file(input_dir, suffix='pth.tar'):
    '''
    Check whether the `input_path` is directory or to the checkpoint file.
        If it is a directory, find the only 'pth.tar' file under it.

    Parameters
    ----------
    input_path: str
        The input path to the pretrained model.
    suffix: 'pth.tar' or 'model'
        The checkpoint file suffix;
        If 'pth.tar', the saved model is a torch model.
        If 'model', the saved model is a scikit-learn based model.
    '''
    suffix = '.' + suffix
    if input_dir.endswith(suffix):
        return input_dir

    ckpt_list = glob.glob(os.path.join(input_dir, '*'+suffix))
    assert len(ckpt_list) <= 1, f'Find more than one checkpoints under the dir {input_dir}, please specify the one to load.'
    assert len(ckpt_list) > 0, f'Do not find any checkpoint under the dir {input_dir}.'
    return ckpt_list[0]


def check_model_config_file(input_dir):
    '''
    Check whether the `input_path` is directory or to the `model_config.json` file.
        If it is a directory, find the only '.json' file under it.

    Parameters
    ----------
    input_path: str
        The input path to the pretrained model.
    '''
    if input_dir.endswith('.json'):
        return input_dir

    if not os.path.isdir(input_dir):
        # if the input_dir is the given checkpoint model path,
        # we need to find the config file under the same dir.
        input_dir = os.path.dirname(input_dir)

    ckpt_list = glob.glob(os.path.join(input_dir, '*.json'))

    if len(ckpt_list) == 0:
        return None

    # find model_config.json under this input_dir
    model_config_name = [config for config in ckpt_list if 'promptehr_config.json' in config]
    if len(model_config_name) == 1:
        return model_config_name[0]

    # if no model_config.json found, retrieve the only .json file.
    assert len(ckpt_list) <= 1, f'Find more than one config .json under the dir {input_dir}.'
    return ckpt_list[0]


def check_tokenizer_file(input_dir):
    return os.path.join(input_dir,'data_tokenizer.pkl'), os.path.join(input_dir,'model_tokenizer.pkl')



import pdb

import torch
from torch import nn

from pytrial.utils.check import (
    check_checkpoint_file, check_model_config_file, make_dir_if_not_exist
)
from .base import SequenceIndivBase
from ..trainer import IndivSeqTrainer

class LocationAttention(nn.Module):

    def __init__(self, hidden_size, device):
        super(LocationAttention, self).__init__()
        self.hidden_size = hidden_size
        self.attention_value_ori_func = nn.Linear(self.hidden_size, 1)
        self.device = device

    def forward(self, input_data):
        # shape of input_data: <n_batch, n_seq, hidden_size>         
        n_batch, n_seq, hidden_size = input_data.shape
        # shape of reshape_feat: <n_batch*n_seq, hidden_size>       
        reshape_feat = input_data.reshape(n_batch*n_seq, hidden_size)
        # shape of attention_value_ori: <n_batch*n_seq, 1>       
        attention_value_ori = torch.exp(self.attention_value_ori_func(reshape_feat))
        # shape of attention_value_format: <n_batch, 1, n_seq>       
        attention_value_format = attention_value_ori.reshape(n_batch, n_seq).unsqueeze(1)        
        # shape of ensemble flag format: <1, n_seq, n_seq> 
        # if n_seq = 3, ensemble_flag_format can get below flag data
        #  [[[ 0  0  0 
        #      1  0  0
        #      1  1  0 ]]]
        ensemble_flag_format = torch.triu(torch.ones([n_seq, n_seq]), diagonal = 1).permute(1, 0).unsqueeze(0).to(self.device)
        # shape of accumulate_attention_value: <n_batch, n_seq, 1>
        accumulate_attention_value = torch.sum(attention_value_format * ensemble_flag_format, -1).unsqueeze(-1) + 1e-9
        # shape of each_attention_value: <n_batch, n_seq, n_seq>
        each_attention_value = attention_value_format * ensemble_flag_format
        # shape of attention_weight_format: <n_batch, n_seq, n_seq>
        attention_weight_format = each_attention_value/accumulate_attention_value
        # shape of _extend_attention_weight_format: <n_batch, n_seq, n_seq, 1>
        _extend_attention_weight_format = attention_weight_format.unsqueeze(-1)
        # shape of _extend_input_data: <n_batch, 1, n_seq, hidden_size>
        _extend_input_data = input_data.unsqueeze(1)
        # shape of _weighted_input_data: <n_batch, n_seq, n_seq, hidden_size>
        _weighted_input_data = _extend_attention_weight_format * _extend_input_data
        # shape of weighted_output: <n_batch, n_seq, hidden_size>
        weighted_output = torch.sum(_weighted_input_data, 2)
        return weighted_output

class GeneralAttention(nn.Module):

    def __init__(self, hidden_size, device):
        super(GeneralAttention, self).__init__()
        self.hidden_size = hidden_size
        self.correlated_value_ori_func = nn.Linear(self.hidden_size, self.hidden_size)
        self.device = device

    def forward(self, input_data):
        # shape of input_data: <n_batch, n_seq, hidden_size>         
        n_batch, n_seq, hidden_size = input_data.shape
        # shape of reshape_feat: <n_batch*n_seq, hidden_size>       
        reshape_feat = input_data.reshape(n_batch*n_seq, hidden_size)
        # shape of correlated_value_ori: <n_batch, n_seq, hidden_size>       
        correlated_value_ori = self.correlated_value_ori_func(reshape_feat).reshape(n_batch, n_seq, hidden_size)
        # shape of _extend_correlated_value_ori: <n_batch, n_seq, 1, hidden_size>   
        _extend_correlated_value_ori = correlated_value_ori.unsqueeze(-2)
        # shape of _extend_input_data: <n_batch, 1, n_seq, hidden_size>   
        _extend_input_data = input_data.unsqueeze(1)
        # shape of _extend_input_data: <n_batch, n_seq, n_seq, hidden_size> 
        _correlat_value = _extend_correlated_value_ori * _extend_input_data
        # shape of attention_value_format: <n_batch, n_seq, n_seq>       
        attention_value_format = torch.exp(torch.sum(_correlat_value, dim = -1))
        # shape of ensemble flag format: <1, n_seq, n_seq> 
        # if n_seq = 3, ensemble_flag_format can get below flag data
        #  [[[ 0  0  0 
        #      1  0  0
        #      1  1  0 ]]]
        ensemble_flag_format = torch.triu(torch.ones([n_seq, n_seq]), diagonal = 1).permute(1, 0).unsqueeze(0).to(self.device)
        # shape of accumulate_attention_value: <n_batch, n_seq, 1>
        accumulate_attention_value = torch.sum(attention_value_format * ensemble_flag_format, -1).unsqueeze(-1) + 1e-10
        # shape of each_attention_value: <n_batch, n_seq, n_seq>
        each_attention_value = attention_value_format * ensemble_flag_format
        # shape of attention_weight_format: <n_batch, n_seq, n_seq>
        attention_weight_format = each_attention_value/accumulate_attention_value
        # shape of _extend_attention_weight_format: <n_batch, n_seq, n_seq, 1>
        _extend_attention_weight_format = attention_weight_format.unsqueeze(-1)
        # shape of _extend_input_data: <n_batch, 1, n_seq, hidden_size>
        _extend_input_data = input_data.unsqueeze(1)
        # shape of _weighted_input_data: <n_batch, n_seq, n_seq, hidden_size>
        _weighted_input_data = _extend_attention_weight_format * _extend_input_data
        # shape of weighted_output: <n_batch, n_seq, hidden_size>
        weighted_output = torch.sum(_weighted_input_data, 2)
        return weighted_output

class ConcatenationAttention(nn.Module):
    def __init__(self, hidden_size, attention_dim = 16, device = None):
        super(ConcatenationAttention, self).__init__()
        self.hidden_size = hidden_size
        self.attention_dim = attention_dim        
        self.attention_map_func = nn.Linear(2 * self.hidden_size, self.attention_dim)
        self.activate_func = nn.Tanh()
        self.correlated_value_ori_func = nn.Linear(self.attention_dim, 1)
        self.device = device
				
    def forward(self, input_data):
        # shape of input_data: <n_batch, n_seq, hidden_size>         
        n_batch, n_seq, hidden_size = input_data.shape
        # shape of _extend_input_data: <n_batch, n_seq, 1, hidden_size>       
        _extend_input_data_f = input_data.unsqueeze(-2)
        # shape of _repeat_extend_correlated_value_ori: <n_batch, n_seq, n_seq, hidden_size>   
        _repeat_extend_input_data_f = _extend_input_data_f.repeat(1,1,n_seq,1)
        # shape of _extend_input_data: <n_batch, 1, n_seq, hidden_size>   
        _extend_input_data_b = input_data.unsqueeze(1)
        # shape of _repeat_extend_input_data: <n_batch, n_seq, n_seq, hidden_size>   
        _repeat_extend_input_data_b = _extend_input_data_b.repeat(1,n_seq,1,1)
        # shape of _concate_value: <n_batch, n_seq, n_seq, 2 * hidden_size>           
        _concate_value = torch.cat([_repeat_extend_input_data_f, _repeat_extend_input_data_b], dim = -1)        
        # shape of _correlat_value: <n_batch, n_seq, n_seq> 
        _correlat_value = self.activate_func(self.attention_map_func(_concate_value.reshape(-1, 2 * hidden_size)))
        _correlat_value = self.correlated_value_ori_func(_correlat_value).reshape(n_batch, n_seq, n_seq)
        # shape of attention_value_format: <n_batch, n_seq, n_seq>       
        attention_value_format = torch.exp(_correlat_value)
        # shape of ensemble flag format: <1, n_seq, n_seq> 
        # if n_seq = 3, ensemble_flag_format can get below flag data
        #  [[[ 0  0  0 
        #      1  0  0
        #      1  1  0 ]]]
        ensemble_flag_format = torch.triu(torch.ones([n_seq, n_seq]), diagonal = 1).permute(1, 0).unsqueeze(0).to(self.device)
        # shape of accumulate_attention_value: <n_batch, n_seq, 1>
        accumulate_attention_value = torch.sum(attention_value_format * ensemble_flag_format, -1).unsqueeze(-1) + 1e-10
        # shape of each_attention_value: <n_batch, n_seq, n_seq>
        each_attention_value = attention_value_format * ensemble_flag_format
        # shape of attention_weight_format: <n_batch, n_seq, n_seq>
        attention_weight_format = each_attention_value/accumulate_attention_value
        # shape of _extend_attention_weight_format: <n_batch, n_seq, n_seq, 1>
        _extend_attention_weight_format = attention_weight_format.unsqueeze(-1)
        # shape of _extend_input_data: <n_batch, 1, n_seq, hidden_size>
        _extend_input_data = input_data.unsqueeze(1)
        # shape of _weighted_input_data: <n_batch, n_seq, n_seq, hidden_size>
        _weighted_input_data = _extend_attention_weight_format * _extend_input_data
        # shape of weighted_output: <n_batch, n_seq, hidden_size>
        weighted_output = torch.sum(_weighted_input_data, 2)
        return weighted_output

class BuildModel(nn.Module):
    def __init__(self, 
                 input_size = None,
                 embed_size = 16,
                 hidden_size = 8,
                 output_size = 10,
                 bias = True,
                 dropout = 0.5,
                 batch_first = True,
                 label_size = 1,
                 attention_type = 'location_based',
                 attention_dim = 8,
                 device = None):
        super(BuildModel, self).__init__()
        assert input_size != None and isinstance(input_size, int), 'fill in correct input_size' 
 
        self.input_size = input_size        
        self.embed_size = embed_size
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.label_size = label_size
				
        self.embed_func = nn.Linear(self.input_size, self.embed_size)
        self.rnn_model = nn.GRU(input_size = embed_size,
                                 hidden_size = hidden_size,
                                 bias = bias,
                                 dropout = dropout,
                                 bidirectional = True,
                                 batch_first = batch_first)
        if attention_type == 'location_based':
            self.attention_func = LocationAttention(2*hidden_size, device)
        elif attention_type == 'general':
            self.attention_func = GeneralAttention(2*hidden_size, device)
        elif attention_type == 'concatenation_based':
            self.attention_func = ConcatenationAttention(2*hidden_size, attention_dim = attention_dim, device = device)
        else:
            raise Exception('fill in correct attention_type, [location_based, general, concatenation_based]')
        self.output_func = nn.Linear(4 * hidden_size, self.output_size) 
        self.output_activate = nn.Tanh()
        self.predict_func = nn.Linear(self.output_size, self.label_size)
            
    def forward(self, input_data):
        
        """
        
        Parameters
        
        ----------
        input_data = {
                      'X': shape (batchsize, n_timestep, n_featdim)
                      'M': shape (batchsize, n_timestep)
                      'cur_M': shape (batchsize, n_timestep)
                      'T': shape (batchsize, n_timestep)
                     }
        
        Return
        
        ----------
        
        all_output, shape (batchsize, n_timestep, n_labels)
            
            predict output of each time step
            
        cur_output, shape (batchsize, n_labels)
        
            predict output of last time step
        
        """

        X = input_data['v']
        
        if 'v_lengths' in input_data:
            v_len = input_data['v_lengths']
            mask, cur_mask = self._create_mask_from_visit_length(v_len=v_len)
            mask = mask.to(X.device)
            cur_mask = cur_mask.to(X.device)
        else:
            mask = cur_mask = None

        batchsize, n_timestep, n_orifeatdim = X.shape
        _ori_X = X.view(-1, n_orifeatdim)
        _embed_X = self.embed_func(_ori_X)
        _embed_X = _embed_X.reshape(batchsize, n_timestep, self.embed_size)
        _embed_F, _ = self.rnn_model(_embed_X)
        _embed_F_w = self.attention_func(_embed_F)
        _mix_F = torch.cat([_embed_F, _embed_F_w], dim = -1)
        _mix_F_reshape = _mix_F.view(-1, 4 * self.hidden_size)
        outputs = self.output_activate(self.output_func(_mix_F_reshape)).reshape(batchsize, n_timestep, self.output_size)
        n_batchsize, n_timestep, output_size = outputs.shape

        all_output = self.predict_func(outputs.reshape(n_batchsize*n_timestep, output_size)).\
                         reshape(n_batchsize, n_timestep, self.label_size)

        if mask is not None and cur_mask is not None:
            all_output *= mask.unsqueeze(-1)
            cur_output = (all_output * cur_mask.unsqueeze(-1)).sum(dim=1)
        else:
            cur_output = all_output.sum(1)

        # return all_output, cur_output
        # TODO: all_output is the per visit label prediction, currently we only support patient-level prediction.
        return cur_output

    def _create_mask_from_visit_length(self, v_len):
        mask = torch.zeros(len(v_len),max(v_len))
        cur_mask = torch.zeros(len(v_len),max(v_len))

        for i in range(len(v_len)):
            mask[i, :v_len[i]] = 1
            cur_mask[i, v_len[i]-1] = 1
        return mask, cur_mask

class Dipole(SequenceIndivBase):
    '''
    Implement Dipole for longitudinal patient records predictive modeling [1]_.
    
    Parameters
    ----------
    vocab_size: list[int]
        A list of vocabulary size for different types of events, e.g., for diagnosis, procedure, medication.

    orders: list[str]
        A list of orders when treating inputs events. Should have the same shape of `vocab_size`.

    mode: str
        Prediction traget in ['binary','multiclass','multilabel','regression'].

    output_dim: int
        Output dimension of the model.

        - If binary classification, output_dim=1;
        - If multiclass/multilabel classification, output_dim=n_class
        - If regression, output_dim=1.

    max_visit: int
        The maximum number of visits for input event codes.

    attention_type: {'general', 'concatenation_based', 'location_based'}

        Apply attention mechnism to derive a context vector that captures relevant information to 
        help predict target.

        - 'location_based': Location-based Attention. Alocation-based attention function is to calculate the weights solely from hidden state 
        
        - 'general': General Attention. An easy way to capture the relationship between two hidden states
        
        - 'concatenation_based': Concatenation-based Attention. Via concatenating two hidden states, then use multi-layer perceptron(MLP) to calculatethe contextvector

    attention_dim: int
        It is the latent dimensionality used for attention weight computing just for for concatenation_based attention mechnism 

    emb_size: int
        Embedding size for encoding input event codes.

    hidden_size : int, optional (default = 8)
        The number of features of the hidden state h
 
    hidden_output_size : int, optional (default = 8)
        The number of mix features

    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    weight_decay: float
        Regularization strength for l2 norm; must be a positive float.
        Smaller values specify weaker regularization.
    
    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    num_worker: int
        Number of workers used to do dataloading during training.

    device: str
        The model device.
    
    Notes
    -----
    ..  [1] Ma, F., Chitta, R., Zhou, J., You, Q., Sun, T., & Gao, J. (2017, August). Dipole: Diagnosis prediction in healthcare via attention-based bidirectional recurrent neural networks. In Proceedings of the 23rd ACM SIGKDD international conference on knowledge discovery and data mining (pp. 1903-1911).
    '''
    def __init__(self,
        vocab_size,
        orders,
        mode,
        output_dim=None,
        max_visit=None,
        attention_type='location_based',
        attention_dim=8,
        emb_size=16,
        hidden_size=8,
        hidden_output_size=8,
        dropout=0.5,
        learning_rate=1e-4,
        weight_decay=1e-4,
        batch_size=64,
        epochs=10,
        num_worker=0,
        device='cuda:0',
        experiment_id='test',
        ):
        super().__init__(experiment_id, mode=mode, output_dim=output_dim)
        self.config = {
            'mode':self.mode,
            'vocab_size':vocab_size,
            'max_visit':max_visit,
            'attention_type':attention_type,
            'attention_dim':attention_dim,
            'emb_size':emb_size,
            'hidden_size':hidden_size,
            'hidden_output_size':hidden_output_size,
            'output_dim':self.output_dim,
            'dropout':dropout,
            'device':device,
            'learning_rate':learning_rate,
            'batch_size':batch_size,
            'weight_decay':weight_decay,
            'epochs':epochs,
            'num_worker':num_worker,
            'orders':orders,
            }
        self.config['total_vocab_size'] = sum(vocab_size)
        self.device = device
        self._build_model()
    
    def predict(self, test_data):
        '''
        Predict patient outcomes using longitudinal trial patient sequences.
        
        Parameters
        ----------
        test_data: SequencePatient
            A `SequencePatient` contains patient records where 'v' corresponds to 
            visit sequence of different events.

        '''
        test_dataloader = self.get_test_dataloader(test_data)
        outputs = self._predict_on_dataloader(test_dataloader)
        return outputs['pred']

    def fit(self, train_data, valid_data):
        '''
        Train model with sequential patient records.
        
        Parameters
        ----------
        train_data: SequencePatientBase
            A `SequencePatientBase` contains patient records where 'v' corresponds to 
            visit sequence of different events; 'y' corresponds to labels.

        valid_data: SequencePatientBase
            A `SequencePatientBase` contains patient records used to make early stopping of the
            model.
        '''
        self._input_data_check(train_data)
        if valid_data is not None: self._input_data_check(valid_data)
        self._fit_model(train_data, valid_data=valid_data)

    def load_model(self, checkpoint):
        '''
        Load pretrained model from the disk.

        Parameters
        ----------
        checkpoint: str
            The input directory that stores the trained pytorch model and configuration.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        
        self.model.load_state_dict(state_dict['model'])

    def save_model(self, output_dir):
        '''
        Save the pretrained model to the disk.

        Parameters
        ----------
        output_dir: str
            The output directory that stores the trained pytorch model and configuration.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir
        
        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model.state_dict()}, output_dir=output_dir)

    def _build_model(self):
        self.model = BuildModel(
            input_size=self.config['total_vocab_size'],
            embed_size=self.config['emb_size'],
            hidden_size=self.config['hidden_size'],
            output_size=self.config['hidden_output_size'],
            dropout=self.config['dropout'],
            label_size=self.config['output_dim'],
            attention_type=self.config['attention_type'],
            attention_dim=self.config['attention_dim'],
            device=self.config['device'],
            )
        self.model.to(self.device)

    def _fit_model(self, train_data, valid_data=None):
        test_metric_dict = {
            'binary': 'auc',
            'multiclass': 'acc',
            'regression': 'mse',
            'multilabel': 'f1', # take average of F1 scores
            }
        train_dataloader = self.get_train_dataloader(train_data)
        loss_models = self._build_loss_model()
        train_objectives = [(train_dataloader, loss_model) for loss_model in loss_models]

        if self.config['mode'] == 'regression':
            less_is_better = True
        else:
            less_is_better = False

        trainer = IndivSeqTrainer(model=self,
            train_objectives=train_objectives,
            test_data=valid_data,
            test_metric=test_metric_dict[self.config['mode']],
            less_is_better=less_is_better,
            )
        trainer.train(**self.config)


'''
Implement Gaussian Copula model for tabular simulation
prediction in clinical trials.
'''
import os
import warnings
import joblib
from copulas.multivariate import GaussianMultivariate

from .base import TabularSimulationBase
from pytrial.data.patient_data import TabularPatientBase
from pytrial.utils.check import check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist

warnings.filterwarnings('ignore')


class BuildModel:
    def __new__(self) -> GaussianMultivariate:
        return GaussianMultivariate()

class GaussianCopula(TabularSimulationBase):
    '''
    Implement Gaussian Copula model for tabular patient simulation.

    Parameters
    ----------
    experiment_id: str, optional (default='trial_simulation.tabular.gaussiancopula')
        The name of current experiment. Decide the saved model checkpoint name.
    '''
    def __init__(
        self,
        experiment_id='trial_simulation.tabular.gaussiancopula',
    ) -> None:
        super().__init__(experiment_id=experiment_id)
        
    def fit(self, train_data):
        '''
        Train gaussian copula model to simulate patient outcome
        with tabular input data.
        
        Parameters
        ----------
        train_data: dict or TabularPatientBase
            The training data, which is the real tabular patient data.
        '''
        self._input_data_check(train_data)
        self._build_model()
        if isinstance(train_data, TabularPatientBase):
            dataset = train_data.df
        if isinstance(train_data, dict): 
            dataset = TabularPatientBase(train_data, transform=True)
            dataset = dataset.df
        self._fit_model(dataset)
        self.metadata = train_data.metadata
        self.raw_dataset = train_data

    def predict(self, n=200):
        '''
        simulate a new tabular data with number_of_predictions.

        Parameters
        ----------
        n: int
            The number of synthetic samples to generation.

        Returns
        -------
        ypred: TabularPatientBase
            A new tabular data simulated by the model
        '''
        ypred = self.model.sample(n) # build df
        ypred = self.raw_dataset.reverse_transform(ypred) # transform back
        return ypred # output: dataset, same as the input dataset

    def save_model(self, output_dir=None):
        '''
        Save the learned gaussian copula model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir

        #self._save_config(self.config, output_dir=output_dir)
        ckpt_path = os.path.join(output_dir, 'gaussiancopula.model')
        joblib.dump(self.model, ckpt_path)

    def load_model(self, checkpoint=None):
        '''
        Save the learned gaussian copula model to the disk.

        Parameters
        ----------
        checkpoint: str or None
            The path to the saved model.

            - If a directory, the only checkpoint file `.model` will be loaded.
            - If a filepath, will load from this file;
            - If None, will load from `self.checkout_dir`.
        '''
        if checkpoint is None:
            checkpoint = self.checkout_dir

        checkpoint_filename = check_checkpoint_file(checkpoint, suffix='model')
        config_filename = check_model_config_file(checkpoint)
        self.model = joblib.load(checkpoint_filename)

    def _build_model(self):
        self.model = BuildModel()

    def _fit_model(self, data):
        self.model.fit(data)


import pdb
import torch
import numpy as np 

from pytrial.data.patient_data import SequencePatientBase

def concat_dataset(*datasets):
    '''
    Concatenate two torch.utils.data.Dataset objects.
    At present, this function ONLY supports `pytrial.data.patient_data.SequencePatient` objects.
    It will ignore the part if one of the dataset does not have.
    For example, if one of the datasets does not have `label`, then the output concatenated dataset will not have `label`.

    Parameters
    ----------
    datasets: tuple[torch.utils.data.Dataset]
        A tuple of torch.utils.data.Dataset objects.
    '''
    for dataset in datasets:
        assert isinstance(dataset, SequencePatientBase), '`dataset` must be a `SequencePatient` object.'
    
    # concat the data
    label, visit, feature = [], [], []
    for dataset in datasets:
        visit.extend(dataset.visit)
        if dataset.feature is not None:
            feature.append(dataset.feature)
        if dataset.label is not None:
            label.append(dataset.label)

    data = {'v': visit}

    if len(feature) == len(datasets):
        feature = np.concatenate(feature, axis=0)
        data['x'] = feature
    
    if len(label) == len(datasets):
        label = np.concatenate(label, axis=0)
        data['y'] = label

    cls = SequencePatientBase
    
    new_data = cls(
        data=data,
        metadata=dataset.metadata,
        )
    return new_data

def split_dataset(dataset, split=0.8):
    '''
    Split a dataset into two datasets.

    At present, this function ONLY supports `pytrial.data.patient_data.SequencePatient` objects.
    '''
    assert isinstance(dataset, SequencePatientBase), '`dataset` must be a `SequencePatient` object.'

    cls = SequencePatientBase
    n = len(dataset)
    n1 = int(n * split)
    n2 = n - n1
    data1, data2 = torch.utils.data.random_split(dataset, [n1, n2])
    indice1 = data1.indices
    indice2 = data2.indices

    data1 =  {'v': [dataset.visit[ind] for ind in indice1],}
    
    if dataset.feature is not None:
        data1['x'] = dataset.feature[indice1]

    if dataset.label is not None:
        if isinstance(dataset.label, list):
            dataset.label = np.array(dataset.label)
        data1['y'] = dataset.label[indice1]

    new_data1 = cls(
        data=data1,
        metadata=dataset.metadata,
        )
    
    data2 = {'v': [dataset.visit[ind] for ind in indice2],}
    if dataset.feature is not None:
        data2['x'] = dataset.feature[indice2]

    if dataset.label is not None:
        if isinstance(dataset.label, list):
            dataset.label = np.array(dataset.label)
        data2['y'] = dataset.label[indice2]

    new_data2 = cls(
        data=data2,
        metadata=dataset.metadata,
        )

    return new_data1, new_data2